#include "AST.h"
#include "Worklist.h"
#include "optimize.h"
#include "code-util.h"

static bool assignment(TreeNode *t)
{
  return isExpressionStmtNode(t) && isAssignNode(t->child(0));
}

/* t is an AssignNode.  s is a set of assignments believed to be useless.
   If the result of t is used by a statement not in s then t is useful.
   I.e., if the result of t is used by a non-useless stmt then return false.
   Also return false if t has or may have side-effects. */
static bool useless(TreeNode *t, const treeSet *s)
{
  TreeNode *lhs = t->opnd0();
  if (DEBUG_DCE) {
    cout << "Checking utility of ";
    cout << PCODE(t);
    cout << endl;
  }

  if (!isObjectNode(lhs))
    return false;
  if (!isLocalOrParam(lhs->decl()))
    return false;
  TreeNode *rhs = t->opnd1();
  
  // Do we have to execute rhs because of possible side effects?
  // If so, return false immediately.
  if (!(isObjectNode(rhs) && isLocalOrParam(rhs->decl())) &&
      !isSideEffectFreeAndCannotFail(rhs) &&
      !isConstantOrCastOfConstant(rhs) &&
      !isNullOrCastOfNull(rhs) &&
      !isTrivialPointOrDomain(rhs))
    return false;

  treeSet *rvu = RvalueUses(lhs);
  treeSet *u = exprset_to_stmtset(rvu);
  bool r = isSubset(u, s);
  if (DEBUG_DCE) {
    cout << "Stmts with rvalue uses stemming from ";
    cout << PLCODE(lhs);
    cout << ":" << endl;
    printTreeSet(u, cout, 1);
    cout << "(all uses:" << endl;
    printTreeSet(allUses(lhs), cout, 1);
    cout << ")" << endl << "Therefore, ";
    cout << PCODE(t);
    cout << (r ? " is" : " is not") << " useless" << endl << endl;
  }
  delete rvu;
  delete u;
  return r;
}

static void showAssignmentsAndUses(const treeSet *r, ostream &os)
{
  for (treeSet::const_iterator i = r->begin();
       i != r->end();
       i++) {
    (*i)->pseudoprint(os, 0);
    os << endl << "  Uses of ";
    (*i)->child(0)->child(0)->pseudoprint(os, 0);
    os << endl;
    printTreeSetWithParents(list_to_set((*i)->child(0)->child(0)->getUses()), os, 2);
    os << endl << "  Defs:" << endl;
    printTreeSetWithParents(list_to_set((*i)->child(0)->child(0)->getDefs()), os, 2);
    os << endl;
  }
}

treeSet * TreeNode::uselessAssignments()
{
  Worklist w(this);
  w.filter(assignment);

  treeSet *r = list_to_set(w.asList());

  if (DEBUG_DCE) {
    cout << "Preparing to eliminate useless assignments." << endl << endl;
    showAssignmentsAndUses(r, cout);
    cout << endl;
  }

  bool progress;
  do {
    progress = false;
    for (treeSet::iterator i = r->begin();
	 i != r->end(); )
      if (!useless((*i)->child(0), r)) {
	r->erase(i++);
	progress = true;
      } 
      else i++;
  } while (progress);

  if (DEBUG_DCE) {
    cout << "Useless Assignments:" << endl;
    printTreeSet(r, cout, 1);
  }
  return r;
}

#if 0

void TreeNode::checkIfIterationPointIsNeeded(treeSet *useless)
{
  for (int i = arity(); i-- > 0; )
    child(i)->checkIfIterationPointIsNeeded(useless);
}

#undef DEBUG_DCE
#define DEBUG_DCE 1
void ForEachStmtNode::checkIfIterationPointIsNeeded(treeSet *useless)
{
  if (stmt() == TreeNode::omitted)
    needPoint = false;
  else {
    TreeNode *var = vars()->child(0)->simpName();
    treeSet *uses = list_to_set(var->getUses());
    if (DEBUG_DCE) {
      cout << "Checking uses of foreach iteration point ";
      cout << PLCODE(var);
      cout << endl;
      printTreeSet(uses, cout, 2);
    }
    stmt()->checkIfIterationPointIsNeeded(useless);
    needPoint = !isSubset(uses, useless);
    if (DEBUG_DCE)
      cout << "needPoint = " << needPoint << endl;
    delete uses;
  }
}

#endif
