#include <cassert>
#include "utils.h"
#include "AST.h"
#include "CodeContext.h"
#include "CfSource.h"
#include "code.h"
#include "code-util.h"
#include "code-grid.h"
#include "nullify-local.h"
#include "optimize.h"
#include "lower.h"
#include "version.h"
#include "code-MIVE.h"
#include "Worklist.h"
#include "ullist.h"
#include "cfg.h"

extern int compile_verbose_level;

static TreeNode *optimize(TreeNode *t);

void DummyNode::emitStatement( CodeContext &context ) {}

PrimitiveLitNode *newBoolLit(bool b, SourcePosn p)
{
  return new PrimitiveLitNode(Literal(b), p);
}

TreeNode *CloneTree(TreeNode *t, bool cleanUseDef)
{
  TreeNode *newNode = t->clone();
  if (cleanUseDef) {
    newNode->setUses(NULL);
    newNode->setDefs(NULL);
  }
  const int childCount = newNode->arity();
  for (int sweep = 0; sweep < childCount; ++sweep)
    newNode->child(sweep, CloneTree(newNode->child(sweep), cleanUseDef));

  return newNode;
}

static bool assignablenode(TreeNode *t) {
  return t->hasLval();
}

static bool objectnode(TreeNode *t)
{
  return isObjectNode(t);
}

static bool exprnode(TreeNode *t)
{
  return isExprNode(t);
}

/////////////////////////////////////////////////////////////////////////////
// Foreach Opts 
/////////////////////////////////////////////////////////////////////////////

/* The enclosing loop of a foreach loop, if any */
static map_tree_to_tree enclLoop;

#define REPORT_AND_ADD(x, o)			\
((DEBUG_OPT ?					\
  ((o << "Add "),				\
   (o << PLCODE(x)),				\
   (o << " to worklist"),			\
   (o << endl), 0) :				\
  0),						\
(w)->add((x)))

#define ADD(x, w) do if (isExprNode((x))) REPORT_AND_ADD((x), cout); while (0)

void propagateChanges(Worklist *w, TreeNode *t)
{
  ADD(t->parent(), w);
  foreach (u, llist<TreeNode *>, *(t->getUses()))
    ADD(*u, w);
}

#undef ADD
#undef REPORT_AND_ADD

/* PROPAGATE(w) is used as (lambda (t) (propagateChanges w t)), but since
   we don't have lambda we use current_worklist as a hackish way to store w. */

Worklist *current_worklist;

static void propToCurrentWorklist(TreeNode *t)
{
  propagateChanges(current_worklist, t);
}

#define PROPAGATE(w) \
    ((current_worklist = (w)), propToCurrentWorklist)

static void invAndPropToCurrentWorklist(TreeNode *t, TreeNode *WRTloop)
{
  if (!isInvar(t, WRTloop)) {
    foundInvariant(t, WRTloop);
    propagateChanges(current_worklist, t);
  }
}

/* similar to PROPAGATE(w), above */
#define INVARIANT_AND_PROPAGATE(w) \
    ((current_worklist = (w)), invAndPropToCurrentWorklist)

extern void gotMIVE(TreeNode *t, TreeNode *WRTloop, MIVE *m);

#define GOT_MIVE(w) \
    ((current_worklist = (w)), gotMIVE)

/////////////////////////////////////////////////////////////////////////////
// Figure out if t is invariant and/or if we can put a polynomial on it.
/////////////////////////////////////////////////////////////////////////////

static void computeMIVE(TreeNode *t, TreeNode *WRTloop, treeSet *loopContents,
			Worklist *w) 
{
  /* If t is the child of an assignment, work on t's parent instead. */
  if (isAssignNode(t->parent())) {
    w->add(t->parent());
    return;
  }

  /* If t is already known to be invariant there's no need to
     recompute anything, unless t is an assignment, in which case we
     may have improved MIVE info to deal with. */
  bool recompute_all = !isInvar(t, WRTloop), assign = isAssignNode(t);

  /* Determine and record whether it is an invariant. */
  if (recompute_all) 
    t->loopInvariant(WRTloop, loopContents, INVARIANT_AND_PROPAGATE(w));

  /* Try to compute a polynomial describing it. */
  if (recompute_all || assign)
    t->computeMIVE(WRTloop, GOT_MIVE(w));
}

/* Build a list that contains the given set of nodes. */
llist<TreeNode *> * set_to_list(const treeSet *s)
{
  llist<TreeNode *> * result = NULL;
  for (treeSet::const_iterator i = s->begin(); i != s->end(); i++)
    result = cons((TreeNode *) *i, result);
  return result;
}

size_t size(const treeSet *s)
{
  size_t count = 0;
  if (s != NULL)
    for (treeSet::const_iterator i = s->begin(); i != s->end(); i++)
      count++; 
  return count;
}

/* Build a treeSet that contains the given list of nodes. */
treeSet * list_to_set(llist<TreeNode *> *l)
{
  treeSet *n = new treeSet;
  foreach (t, llist<TreeNode *>, *l)
    n->insert(*t);
  return n;
}

CFGset * list_to_CFGset(llist<CfgNode *> *l)
{
  CFGset *n = new CFGset;
  foreach (t, llist<CfgNode *>, *l) {
    n->insert(*t);
  }
  return n;
}

/* The loopAnal pass tags each foreach loop with a list of loop invariants
  and computes polynomials (over invariants and induction variables) that
  represent what certain exprs are computing.  The workhorse is foreachAnal().
*/

void TreeNode::loopAnal(TreeNode *enclosingLoop)
{
  for (int i = arity(); i-- > 0; )
    child(i)->loopAnal(enclosingLoop);
}

#define pf(m, i) (m)->p[(i)]->asPolyFactor()

/* This will terminate because at each step either the worklist gets
   smaller or the information we have gets better.
*/
static void foreachAnal(ForEachStmtNode *l)
{
  if (DEBUG_LOOP_ANAL) {
    cout << endl << "foreachAnal: " << l->position().asString()
	 << endl << endl;
  }
  Worklist w(l);
  treeSet *loopContents = list_to_set(cons(l->vars()->child(0), w.asList()));
  w.filter(exprnode);
  MIVEcontext *e = MIVEcontextOfLoop(l) = new MIVEcontext();
  TreeNode *pairnode = l->vars()->child(0);
  gotMIVE(pairnode->simpName(), l,
	  e->adjoin(pairnode->simpName()->ident(),
		    pairnode->initExpr()->type()->tiArity(),
		    l));

  while (!w.empty()) {
    TreeNode *t = w.pop();
    if (isExprNode(t)) {
      // if (isAssignNode(t->parent())) continue;

      if (DEBUG_LOOP_ANAL) {
	cout << endl << "Working on: "; cout << PCODE(t); cout << endl << endl;
      }
      computeMIVE(t, l, loopContents, &w);

    } else {
      // should not happen
      cout << "Not an ExprNode: ";
      t->pseudoprint(cout, 2);
      cout << endl;
    }
  }
  delete loopContents;
  
  setDerivs(l);
}

void ForEachStmtNode::loopAnal(TreeNode *enclosingLoop)
{
  if (compile_verbose_level >= 3)
    cout << "   loop analysis of foreach at " << position().asString() << endl;
  enclLoop[this] = enclosingLoop;
  initInvar(this);

  for (int i = arity(); i-- > 0; )
    child(i)->loopAnal(this);

  foreachAnal(this);
}

/////////////////////////////////////////////////////////////////////////////
// utils
/////////////////////////////////////////////////////////////////////////////

// deleteStmts() behavior is documented in AST.h

TreeNode *TreeNode::deleteStmts(treeSet *s)
{
  const int childCount = arity();
  for (int sweep = 0; sweep < childCount; ++sweep)
    child(sweep, child(sweep)->deleteStmts(s));

  return this;
}

TreeNode *BlockNode::deleteStmts(treeSet *s)
{
  TreeNode *l = child(0);
  TreeNode *result = this;
  int c = l->arity();
  bool del = false; // whether any top-level stmts in the block must be deleted

#if 0
  cout << endl << "Deleting stmts from block:" << endl;
  pseudoprint(cout, 0);
  cout << endl;
#endif

  for (int i = 0; i < c; ++i)
    if (contains(s, l->child(i)) ||
	isExpressionStmtNode(l->child(i)) &&
	contains(s, l->child(i)->expr()))
      del = true;
    else
      l->child(i, l->child(i)->deleteStmts(s));
  
  // If top-level stmts must be deleted, then rebuild the block from scratch.
  if (del) {
    llist<TreeNode *> *q = NULL;
    while (c-- > 0)
      if (!(contains(s, l->child(c)) ||
	    isExpressionStmtNode(l->child(c)) &&
	    contains(s, l->child(c)->expr())))
	q = cons(l->child(c), q);
    result = q ? (TreeNode *)new BlockNode(q, position()) : (TreeNode *)new EmptyStmtNode();
  }

#if 0
  cout << endl << "Result of deleting stmts from block:" << endl;
  result->pseudoprint(cout, 0);
  cout << endl;
#endif

  return result;
}

TreeNode *StatementNode::deleteStmts(treeSet *s)
{
  if (contains(s, this))
    return new EmptyStmtNode();
  else
    return TreeNode::deleteStmts(s);
}

TreeNode *ExpressionStmtNode::deleteStmts(treeSet *s)
{
  if (contains(s, expr()))
    return new EmptyStmtNode();
  else
    return TreeNode::deleteStmts(s);
}

/////////////////////////////////////////////////////////////////////////////

treeSet *allDefs(treeSet *vars)
{
  treeSet *result = new treeSet();
  for (treeSet::iterator i = vars->begin(); i != vars->end(); i++) {
    treeSet *s = list_to_set(((TreeNode *) (*i))->getDefs());
    destructively_union(result, s);
    delete s;
  }
  return result;
}

treeSet *allUses(TreeNode *t)
{
  return list_to_set(t->getUses());
}

static bool opnd0_of_assignnode(TreeNode *t)
{
  TreeNode *p = t->parent();
  return (isAssignNode(p) && p->opnd0() == t);
}

static bool not_opnd0_of_assignnode(TreeNode *t)
{
  bool result = !opnd0_of_assignnode(t);
  if (DEBUG_OPT) {
    if (!result)
      cout << "Not ";
    cout << "Rvalue: ";
    PCODE(t);
    cout << " (because parent is ";
    PCODE(t->parent());
    cout << ")" << endl;
  }
  return result;
}

// Return a set of uses that contains all Lvalue uses.
treeSet *LvalueUses(TreeNode *t)
{
  return list_to_set(filter(opnd0_of_assignnode, t->getUses()));
}

// Return a set of uses that contains all Rvalue uses.
treeSet *RvalueUses(TreeNode *t)
{
  llist<TreeNode *> *l = filter(not_opnd0_of_assignnode, t->getUses());
  treeSet *s =  list_to_set(l);
  free_all(l);
  return s;
}

treeSet *ObjectNodes(TreeNode *e)
{
  Worklist w(e);
  return list_to_set(w.filter(objectnode)->asList());
}

treeSet *AssignableNodes(TreeNode *e)
{
  Worklist w(e);
  return list_to_set(w.filter(assignablenode)->asList());
}


/////////////////////////////////////////////////////////////////////////////
// treeSet / CFGset utils
/////////////////////////////////////////////////////////////////////////////

/* True iff s is a set containing just e. */
bool isSingletonContaining(TreeNode *e, treeSet *s)
{
  treeSet::iterator i = s->begin();
  if (i != s->end() && *i == e) {
    i++;
    return (i == s->end());
  } else
    return false;
}

treeSet *exprset_to_stmtset(treeSet *t)
{
  treeSet * result = new treeSet;
  for (treeSet::iterator i = t->begin(); i != t->end(); i++)
    result = adjoin(result, enclosingStmt(*i));
  return result;
}

treeSet *convert_cfguset_to_treeset(const CFGuset *s)
{
  treeSet *result = new treeSet;
  for (CFGuset::const_iterator i = s->begin(); i != s->end(); i++)
    adjoin(result, (*i)->astNode());
  return result;
}

map_tree_to_treeSet *
destructive_convert_cfgmap_to_astmap(map_CFGnode_to_CFGuset *r)
{
  map_tree_to_treeSet *result = new map_tree_to_treeSet;
  for (map_CFGnode_to_CFGuset::iterator i = r->begin(); i != r->end(); i++)
    (*result)[(*i).first->astNode()] =
      convert_cfguset_to_treeset((*i).second);
  delete r;
  return result;
}

treeSet *destructive_convert_cfgset_to_treeset(CFGset *s)
{
  treeSet *result = new treeSet;
  for (CFGset::iterator i = s->begin(); i != s->end(); i++)
    adjoin(result, (*i)->astNode());
  delete s;
  return result;
}

map_tree_to_treeSet *
destructive_convert_cfgmap_to_astmap(map_CFGnode_to_CFGset *r)
{
  map_tree_to_treeSet *result = new map_tree_to_treeSet;
  for (map_CFGnode_to_CFGset::iterator i = r->begin(); i != r->end(); i++)
    (*result)[(*i).first->astNode()] =
      destructive_convert_cfgset_to_treeset((*i).second);
  delete r;
  return result;
}

// Add t to s if it isn't already there.
treeSet *treeSetAdjoin(treeSet *s, const TreeNode *t)
{
  s->insert(t);
  return s;
}

// Add t to s if it isn't already there.
treeSet *adjoin(treeSet *s, const TreeNode *t)
{
  s->insert(t);
  return s;
}

// Add t to s if it isn't already there.
CFGset *adjoin(CFGset *s, const CfgNode *t)
{
  s->insert((CfgNode *) t);
  return s;
}

bool contains(const CFGset *s, const CfgNode *t)
{
  return s->count((CfgNode *) t);
}


bool contains(const treeSet *s, const TreeNode *t)
{
  return treeSetContains(s, t);
}

bool contains(const treeSet &s, const TreeNode *t)
{
  return treeSetContains(s, t);
}

bool treeSetContains(const treeSet *s, const TreeNode *t)
{
  return s != NULL && s->count(t);
}

bool treeSetContains(const treeSet &s, const TreeNode *t)
{
  return s.count(t);
}

treeSet *copy(treeSet *s)
{
  return destructively_union(new treeSet, s);
}

// Change s to s intersect t.
treeSet *destructively_intersect(treeSet *s, treeSet *t)
{
  treeSet::iterator i = s->begin();
  treeSet::iterator j = t->begin();
  while (true) {
    while (i != s->end() && (j == t->end() || *i < *j))
      s->erase(i++);
    if (i == s->end()) break;
    if (*i == *j)
      i++;
    j++;
  }
  return s;
}

// Change s to s intersect t.
CFGset *destructively_intersect(CFGset *s, CFGset *t)
{
  CFGset::iterator i = s->begin();
  CFGset::iterator j = t->begin();
  while (true) {
    while (i != s->end() && (j == t->end() || *i < *j))
      s->erase(i++);
    if (i == s->end()) break;
    if (*i == *j)
      i++;
    j++;
  }
  return s;
}

// Change s to s union t.
treeSet *destructively_union(treeSet *s, treeSet *t)
{
  for (treeSet::iterator i = t->begin(); i != t->end(); i++)
    treeSetAdjoin(s, (TreeNode *) *i);
  return s;
}

treeSet *filter(bool f(TreeNode *t), const treeSet *s)
{
  treeSet *result = new treeSet;
  for (treeSet::const_iterator i = s->begin(); i != s->end(); i++) {
    TreeNode *t = (TreeNode *) *i;
    if (f(t))
      treeSetAdjoin(result, t);
  }
  return result;
}

treeSet *destructively_filter(bool f(TreeNode *t), treeSet *s)
{
  for (treeSet::iterator i = s->begin(); i != s->end(); ) {
    TreeNode *t = (TreeNode *) *i;
    if (f(t)) i++;
    else s->erase(i++);
  }
  return s;
}

void printTreeSet(const treeSet *s, ostream &os, int m)
{
  if (s == NULL) {
    os << "NULL" << endl;
  } else if (s->empty()) {
    os << "empty" << endl;
  } else
    for (treeSet::const_iterator i = s->begin(); i != s->end(); i++) {
      indent(os, m);
      os << PLCODE2(os, *i) << endl;
    }
}

void printTreeSetWithParents(const treeSet *s, ostream &os, int m)
{
  if (s == NULL) {
    os << "NULL" << endl;
  } else if (s->empty()) {
    os << "empty" << endl;
  } else
    for (treeSet::const_iterator i = s->begin(); i != s->end(); i++) {
      indent(os, m);
      os << PLCODE2(os, *i);
      os << endl;
      indent(os, m + 1);
      os << "Par: ";
      os << PLCODE2(os, (*i)->parent());
      os << endl;
    }
}

// Is s a subset of t?
bool isSubset(const treeSet *s, const treeSet *t)
{
  return isSubset(*s, *t);
}

// Is s a subset of t?
bool isSubset(const treeSet &s, const treeSet &t)
{
  for (treeSet::const_iterator i = s.begin(); i != s.end(); i++) {
    if (!treeSetContains(t, *i)) return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////////

/* Move all VarDeclNodes in the block to the front, and make their
   initializers absent by creating separate assignment statements as needed. */
BlockNode *SortVarDeclsInBlock(BlockNode *b)
{
  TreeListNode *s = b->stmts();
  llist<TreeNode *> *decls = NULL, *newStmts = NULL;
  int c = s->arity();
  for (int i = c; i-- > 0; ) {
    TreeNode *t = s->child(i);
    if (isVarDeclNode(t)) {
      VarDeclNode *d = (VarDeclNode *) t;

      if (!d->initExpr()->absent()) {

	TreeNode *init = d->initExpr();
	d->initExpr(TreeNode::omitted);

	TreeNode *varNode = new ObjectNode (d->simpName()->clone());
	TreeNode *assn = new ExpressionStmtNode(new AssignNode(varNode, init));
	// cout << "Created initializing assignment: ";
	// cout << PCODE(assn->child(0));
	// cout << endl;
	newStmts = cons(assn, newStmts);
      }
      decls = cons(t, decls);
    } else {
      newStmts = cons(t, newStmts);
    }
  }
  newStmts = extend(decls, newStmts);
  b->stmts(new TreeListNode(newStmts));
  return b;
}

SwitchBranchNode *SortVarDeclsInSwitch(SwitchBranchNode *b)
{
  TreeListNode *s = b->stmts();
  llist<TreeNode *> *decls = NULL, *newStmts = NULL;
  int c = s->arity();
  for (int i = c; i-- > 0; ) {
    TreeNode *t = s->child(i);
    if (isVarDeclNode(t)) {
      VarDeclNode *d = (VarDeclNode *) t;

      if (!d->initExpr()->absent()) {

	TreeNode *init = d->initExpr();
	d->initExpr(TreeNode::omitted);

	TreeNode *varNode = new ObjectNode (d->simpName()->clone());
	TreeNode *assn = new ExpressionStmtNode(new AssignNode(varNode, init));
	newStmts = cons(assn, newStmts);
      }
      decls = cons(t, decls);
    } else {
      newStmts = cons(t, newStmts);
    }
  }
  newStmts = extend(decls, newStmts);
  b->stmts(new TreeListNode(newStmts));
  return b;
}

/* def appears on the list returned by use->getDefs().  Return a node
   that has the same run-time value as use, if possible.  Otherwise
   return NULL. */
TreeNode *valueTransferred(TreeNode *def, TreeNode *use)
{
  TreeNode *r = NULL;

  if (isAssignNode(def))
    r = def->opnd0();
  else if (isForEachPairNode(def))
    r = def->simpName();

  if (DEBUG_OPT) {
    cout << "Value transferred from ";
    def->pseudoprint(cout, 0);
    cout << " to ";
    use->pseudoprint(cout, 0);
    cout << ": ";
    if (r == NULL)
      cout << "?";
    else
      r->pseudoprint(cout, 0);
    cout << endl;
  }

  return r;
}

llist<TreeNode *> *add_descendants_to_list(llist<TreeNode *> *l, TreeNode *t)
{
  l = cons(t, l);
  for (int i = t->arity(); i-- > 0; )
    l = add_descendants_to_list(l, t->child(i));
  return l;
}

/* Does s contain any elt of l? */
bool set_contains_any(treeSet *s, llist<TreeNode *> *l)
{
  foreach (u, llist<TreeNode *>, *l)
    if (s->count(*u) > 0) return true;

  return false;
}

/* True iff no DEF of t appears in the set s. */
bool noDefsFromSet(TreeNode *t, treeSet *s)
{
  return !set_contains_any(s, t->getDefs());
}



/////////////////////////////////////////////////////////////////////////////

static treeSet * validCFG = NULL;

// If method does not have a valid CFG then do some analysis now.
// Analyses invoked: basic blocks, use/def, and,
// if dominators is true, foreach dominators.
static void staticAnal(TreeNode *method, bool dominators,
		       bool ignoreNSE = false)
{
  if (validCFG == NULL)
    validCFG = new treeSet;
  if (treeSetContains(validCFG, method))
    return;
  validCFG->insert(method);

  // cout << "Setting up bblocks" << endl;
  setupBblocksMethod(method);
  // cout << "Setting up usedef" << endl;
  setupUdMethod(method);
  if (dominators)
    method->body()->computeForeachDominators(ignoreNSE);
}

static void invalidateCFG()
{
  delete validCFG;
  validCFG = NULL;
}

static void redoDefUse(TreeNode *t)
{
  staticAnal(enclosingMethod(t), false);
}

/////////////////////////////////////////////////////////////////////////////

/* X should be one of PAR's children.  Make Y be PAR's child in X's place. */
static void replaceChild(TreeNode *par, TreeNode *x, TreeNode *y)
{
  for (int i = par->arity(); i-- > 0; )
    if (par->child(i) == x) {
      par->child(i, y);
      return;
    }
  assert(0);
}

static void removeLazyOptNode(TreeNode *par, TreeNode *t)
{
  replaceChild(par, t, t->body());
}

void LazyOptimizeNode::codeGen(CodeContext &context)
{
  TreeNode *t;
  removeLazyOptNode(parent(), this);
#if 1
  staticAnal(enclosingMethod(this), true);
  replaceChild(parent(), body(), t = optimize(body()));
#else
  t = body();
#endif
  if (DEBUG_OPT) {
    cout << "codeGen():\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }
  t->checkIF(true);
  t->codeGen(context);
}

const string LazyOptimizeNode::emitExpression(CodeContext &context)
{
  TreeNode *t;
  removeLazyOptNode(parent(), this);
  staticAnal(enclosingMethod(this), true);
  replaceChild(parent(), body(), t = optimize(body()));
  if (DEBUG_OPT) {
    cout << "emitExpression():\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }
  t->checkIF(true);
  const string result = t->emitExpression(context);
  return result;
}

/////////////////////////////////////////////////////////////////////////////
// Mark parts of the tree to be optimized later.  Optimizations that require
// interprocedural information are done separately.  The LazyOptimizeNodes
// trigger opts during the codegen pass; they are also removed at that time.
/////////////////////////////////////////////////////////////////////////////

void lazyOptimize()
{
  foreach (f, llist<CompileUnitNode *>, *allFiles)
    (*f)->lazyOptimize();
}

static TreeNode *lazyOptNode(TreeNode *s)
{
  if (s == NULL || s->absent())
    return s;
  else
    return new LazyOptimizeNode(s, s->position());
}

TreeNode *TreeNode::lazyOptimize()
{
  for (int i = arity(); i-- > 0; )
    child(i, child(i)->lazyOptimize());
  return this;
}

TreeNode *MethodDeclNode::lazyOptimize()
{
  body(lazyOptNode(body()));
  return this;
}

TreeNode *ConstructorDeclNode::lazyOptimize()
{
  body(lazyOptNode(body()));
  // DOB: optimization has been disabled for constructorCall, because performing DCE there
  // currently causes live expressions to get thrown away
  // I think it's because the method def-use analysis does not include 
  // analysis of uses in constructorCall, therefore it misses those uses and axes the defs
  // there generally isn't much optimization opportunity here anyhow, so it's probably not worth fixing
  //constructorCall(lazyOptNode(constructorCall()));
  return this;
}

TreeNode *VarDeclNode::lazyOptimize()
{
  initExpr(lazyOptNode(initExpr()));
  return this;
}

/////////////////////////////////////////////////////////////////////////////

TreeNode *DCE(TreeNode *t)
{
  redoDefUse(t);
  treeSet *useless = t->uselessAssignments();
  TreeNode *result = t->deleteStmts(useless);
  if (DEBUG_DCE) {
    cout << endl << "After DCE:" << endl;
    result->pseudoprint(cout, 0);
    cout << endl;
  }
  delete useless;
  return result;
}

static TreeNode * foreachOpt(TreeNode *t)
{
  if (DEBUG_OPT) {
    cout << endl;
    cout << "foreachOpt:";
    PLCODE(t);
    cout << endl;
  }
  if (opt_lift || opt_sr || opt_osr) {
    t->loopAnal((TreeNode *) 0);

    /* this next line still converts foreach to foreach+ even if !opt_lift */
    t = t->liftInvariantExprs();

    if (opt_sr) {
      invalidateCFG();
      staticAnal(enclosingMethod(t), true, true);
      t = t->strengthReduce();
    }
  }
  if (DEBUG_OPT) {
    cout << endl;
    cout << "Result of foreachOpt:";
    PLCODE(t);
    cout << endl;
  }
  return t;
}

/*
   Given a piece of code to be optimized, optimize() returns an
   optimized version of that code.  The input may be destructively
   modified in the process.  (Which optimizations to perform is controlled
   by command line flags.)
*/
static TreeNode *optimize(TreeNode *t)
{
  if (DEBUG_OPT) {
    cout << "Optimizing:\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }

  t = foreachOpt(t);

  if (opt_dce) {
    invalidateCFG();
    t->fixParent();
    t = DCE(t);
  }
  
  invalidateCFG();
  t->fixParent();

  if (DEBUG_OPT) {
    cout << "Result of optimization:\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }

  return t;
}
