/*  $Archive:: /Ti/AMUDP/amudp.h                                          $
 *     $Date: Thu, 26 Apr 2001 04:38:13 -0700 $
 * $Revision: 1.4 $
 * Description: AMUDP Header
 * Copyright 2000, Dan Bonachea <bonachea@cs.berkeley.edu>
 */

#ifndef __AMUDP_H
#define __AMUDP_H

#ifdef UFXP
  #include <ufxp.h>
#else
  #include <socket.h>
#endif

#ifdef WIN32
  typedef __int8             int8_t;
  typedef unsigned __int8   uint8_t;
  typedef __int16           int16_t;
  typedef unsigned __int16 uint16_t;
  typedef __int32           int32_t;
  typedef unsigned __int32 uint32_t;
  typedef __int64           int64_t;
  typedef unsigned __int64 uint64_t;
#else
  #include <inttypes.h>
#endif

#include <stdio.h> /* FILE* */

/* miscellaneous macro helpers */
#define _STRINGIFY_HELPER(x) #x
#define _STRINGIFY(x) _STRINGIFY_HELPER(x)

#define AMUDP_LIBRARY_VERSION      1.9
#define AMUDP_LIBRARY_VERSION_STR  _STRINGIFY(AMUDP_LIBRARY_VERSION)

/* naming policy:
  AM-defined things start with AM_
  internal things start with amudp_ or AMUDP_
  */

/* ------------------------------------------------------------------------------------ */
/* Internal constants */
#define AMUDP_MAX_SHORT    8     /* max number of handler arguments, >=8 */
#define AMUDP_MAX_MEDIUM   512   /* max. data transmission unit for medium messages, >= 512 */
#ifdef UFXP
#define AMUDP_MAX_LONG     (AMUDP_MAX_MEDIUM*256)  /* max. data size for xfer and get operations >= 8192 */
#else
#define AMUDP_MAX_LONG     65000  /* max. UDP datagram */
#endif

#define AMUDP_MAX_NUMHANDLERS      256  /* max. handler-table entries >= 256 */
#define AMUDP_MAX_NUMTRANSLATIONS  256  /* max. translation-table entries >= 256 */
#define AMUDP_MAX_SEGLENGTH  1073741824 /* max. dest_offset */

#define AMUDP_MAX_BUNDLES          255  /* max bundles that can be allocated */
#define AMUDP_MAX_NETWORKDEPTH     1024 /* max depth we ever allow user to ask for */
#define AMUDP_MAX_SPMDPROCS        AMUDP_MAX_NUMTRANSLATIONS  /* max SPMD procs we support */

#define AMUDP_COLLECT_LATENCY_STATS   1
/* ------------------------------------------------------------------------------------ */
/* Simple user-visible types */

/* Endpoint tag */
typedef uint64_t tag_t;

/* Handler index */
typedef uint8_t handler_t;

/* Endpoint name */
#ifdef UFXP
  typedef ufxp_addr_t en_t;
#else
  typedef struct sockaddr_in en_t;
#endif

struct amudp_ep; /* forward decls */
struct amudp_buf;

/* ------------------------------------------------------------------------------------ */
/* Internal types */

/* message flags */
 /* 0-1: category
  * 2:   request vs. reply 
  * 3:   sequence number
  * 4-7: numargs
  */
typedef unsigned char amudp_flag_t;
typedef enum {
  amudp_Short=0, 
  amudp_Medium=1, 
  amudp_Long=2,
  amudp_NumCategories=3
  } amudp_category_t;

#define AMUDP_MSG_SETFLAGS(pmsg, isreq, cat, numargs, seqnum) \
  ((pmsg)->flags = (amudp_flag_t) (                           \
                   (((numargs) & 0xF) << 4)                   \
                 | (((seqnum) & 0x1) << 3)                    \
                 | (((isreq) & 0x1) << 2)                     \
                 |  ((cat) & 0x3)                             \
                   ))
#define AMUDP_MSG_NUMARGS(pmsg)   ( ( ((unsigned char)(pmsg)->flags) >> 4 ) & 0xF)
#define AMUDP_MSG_SEQNUM(pmsg)    (!!(((unsigned char)(pmsg)->flags) & 0x8))
#define AMUDP_MSG_ISREQUEST(pmsg) (!!(((unsigned char)(pmsg)->flags) & 0x4))
#define AMUDP_MSG_CATEGORY(pmsg)  ((amudp_category_t)((pmsg)->flags & 0x3))

/* active message header & meta info fields */
typedef struct {
  tag_t         tag;

  uint16_t      instance;
  amudp_flag_t  flags;
  handler_t     handlerId;

  uint16_t      nBytes;
  uint8_t       systemMessageType;
  uint8_t       systemMessageArg;

  uint32_t	destOffset;
  } amudp_msg_t;

/* non-transmitted amudp buffer bookkeeping info -
 * this data must be kept to a bare minimum because it constrains packet size 
 */
typedef struct {
  int8_t handlerRunning;
  int8_t replyIssued;
  uint8_t sourceId;       /* 0-based endpoint id of remote */
  en_t sourceAddr;        /* address of remote */
  struct amudp_ep *dest;  /* ep_t of endpoint that received this message */
  struct amudp_buf *bulkBuffer; /* if non-NULL, points to a bulk buffer 
                              holding the transmitted data fields for this buffer */
  } amudp_bufstatus_t;

/* active message buffer, including message and space for data payload */
typedef struct amudp_buf {
  #ifdef UFXP
    char ufxp_header[UFXP_HEADERLENGTH];
  #endif

  amudp_msg_t	Msg;
  uint8_t     _Data[(4*AMUDP_MAX_SHORT)+AMUDP_MAX_MEDIUM]; /* holds args and data */

  /* received requests & replies only 
    (NOT valid inside bulk buffers, and only the bulkBuffer field is valid in send buffers) */
  amudp_bufstatus_t status;
  } amudp_buf_t;

#define AMUDP_MIN_NETWORK_MSG ((int)&((amudp_buf_t *)NULL)->_Data[0])
#define AMUDP_MAX_NETWORK_MSG ((int)&((amudp_buf_t *)NULL)->_Data[(4*AMUDP_MAX_SHORT)+AMUDP_MAX_MEDIUM])
#define AMUDP_MAXBULK_NETWORK_MSG ((int)&((amudp_buf_t *)NULL)->_Data[(4*AMUDP_MAX_SHORT)+AMUDP_MAX_LONG])

/* message buffer descriptor */
typedef struct {
  int64_t timestamp; 
  #ifdef AMUDP_COLLECT_LATENCY_STATS
    int64_t firstSendTime; /* for statistical purposes only */
  #endif
  uint8_t retryCount;
  uint8_t inuse;
  uint8_t seqNum; /* seq number for next message to be sent/recv'd on this desc */
  } amudp_bufdesc_t;

/* ------------------------------------------------------------------------------------ */
/* Complex user-visible types */

/* statistical collection 
 *  changes here need to also be reflected in the initialization vector AMUDP_initial_stats
 */
typedef struct {
  uint32_t RequestsSent[amudp_NumCategories];
  uint32_t RepliesSent[amudp_NumCategories];
  uint32_t RequestsRetransmitted[amudp_NumCategories];
  uint32_t RepliesRetransmitted[amudp_NumCategories];
  uint32_t RequestsReceived[amudp_NumCategories];   // includes retransmits
  uint32_t RepliesReceived[amudp_NumCategories];    // includes retransmits
  uint32_t ReturnedMessages;
  uint64_t RequestMinLatency;  /* only if AMUDP_COLLECT_LATENCY_STATS */
  uint64_t RequestMaxLatency;  /* only if AMUDP_COLLECT_LATENCY_STATS */
  uint64_t RequestSumLatency;  /* only if AMUDP_COLLECT_LATENCY_STATS */
  uint64_t DataBytesSent[amudp_NumCategories];  /* total of args + data payload for all req/rep, not including retrans */
  uint64_t TotalBytesSent; /* total user level packet sizes for all req/rep, including retrans */
  } amudp_stats_t;

typedef void (*amudp_handler_fn_t)();  /* prototype for handler function */
typedef struct {
  char inuse; // entry in use
  en_t name;  // remote address
  tag_t tag;  // remote tag
  uint8_t id; // id in compressed table
  } amudp_translation_t;

typedef struct {
  uint32_t minDestOffset; /* smallest destOffset seen in this xfer */
  uint32_t runningLength; /* number of bytes copied thus far */
  uint8_t  packetsRemaining; /* number of packets left to be recieved (0 = notinuse)*/
  } bulkslot_t;

typedef struct {
  uint16_t  instanceHint; /* instance hint pointer for request buffer allocation */
  en_t      remoteName;   /* gives us a compacted version of the translation table */
  tag_t     tag;
  bulkslot_t inboundBulkSlot[16]; /* slots for maintaining inbound bulk transfer status */
  } amudp_perproc_info_t;

/* Endpoint bundle object */
typedef struct amudp_eb {
  struct amudp_ep **endpoints;   /* dynamically-grown array of endpoints in bundle */
  int	  n_endpoints;           /* Number of EPs in the bundle */
  int	  cursize;               /* size of the array */
  uint8_t event_mask;            /* Event Mask for blocking ops */
  } *eb_t;

/* Endpoint object */
typedef struct amudp_ep {
  en_t name;            /* Endpoint name */
  tag_t tag;            /* current tag */
  eb_t eb;              /* Bundle of endpoint */

  void *segAddr;        /* Start address of EP VM segment */
  uint32_t segLength;   /* Length of EP VM segment    */

  amudp_translation_t translation[AMUDP_MAX_NUMTRANSLATIONS]; /* translation table */
  amudp_handler_fn_t  handler[AMUDP_MAX_NUMHANDLERS]; /* handler table */

  /* internal structures */
  #ifndef UFXP 
    SOCKET s; /* UDP socket */
    int socketRecvBufferSize; /* only used if USE_SOCKET_RECVBUFFER_GROW */
    int socketRecvBufferMaxedOut;
  #endif
  int P;     /* the number of endpoints we communicate with - also number of translations currently in use */
  int depth; /* network depth, -1 until AM_SetExpectedResources is called */

  /* buffer descriptor tables */
  amudp_bufdesc_t* requestDesc;
  amudp_bufdesc_t* replyDesc;

  int timeoutCheckPosn; /* current position of the timeout-checking circular walk */

  amudp_perproc_info_t *perProcInfo; 

  /* buffers for outbound messages */
  amudp_buf_t* requestBuf;
  amudp_buf_t* replyBuf;
  amudp_buf_t* temporaryBuf; /* a single extra buffer used for temporary operations */

  /* buffers for inbound messages */
  #ifndef UFXP
    /* invariant: buffer empty when rxFreeIdx == rxReadyIdx
     *            buffer full when (rxFreeIdx + 1) % rxNumBufs == rxReadyIdx
     * active buffers are rxReadyIdx ... rxFreeIdx-1 (with circular wrap-around)
     */
    amudp_buf_t* rxBuf;  /* circular buffer */
    uint32_t rxNumBufs;  /* size of circular buffer */
    uint32_t rxReadyIdx; /* oldest unserviced, received message */
    uint32_t rxFreeIdx;  /* beginning of free list */
  #endif

  amudp_stats_t stats;  /* statistical collection */

  amudp_buf_t **bulkBufferPool; /* list of all bulk buffers */
  int bulkBufferPoolSz;     /* length of list */
  int bulkBufferPoolFreeCnt; /* number of those that are free (which appear first in list) */

  } *ep_t;

/* ------------------------------------------------------------------------------------ */
/* User-visible constants */

#define AM_ALL     1    /* Deliver all messages to endpoint */
#define AM_NONE    0    /* Deliver no messages to endpoint */

typedef enum {
  AM_NOEVENTS,   /* No endpoint state transition generates an event */
  AM_NOTEMPTY,   /* A nonempty receive pool or a receive pool that has 
                    a message delivered to it generates an event */
  /* AM_CANSEND, */ /* TODO: can send without blocking */
  AM_NUMEVENTMASKS
  } amudp_eventmask_t;

typedef enum {
    AM_SEQ,             /* Sequential bundle/endpoint access */
    AM_PAR,             /* Concurrent bundle/endpoint access */
    AM_NUM_BUNDLE_MODES
} amudp_bundle_mode_t;

/*
 * Return values to Active Message and Endpoint/Bundle API functions
 */
#define AM_OK           0       /* Function completed successfully */
#define AM_ERR_NOT_INIT 1       /* Active message layer not initialized */
#define AM_ERR_BAD_ARG  2       /* Invalid function parameter passed */
#define AM_ERR_RESOURCE 3       /* Problem with requested resource */
#define AM_ERR_NOT_SENT 4       /* Synchronous message not sent */
#define AM_ERR_IN_USE   5       /* Resource currently in use */

/*
 * Error codes for the AM error handler (status).
 */
#define EBADARGS          1     /* Arguments to request or reply function invalid    */
#define EBADENTRY         2     /* X-lation table index selected unbound table entry */
#define EBADTAG           3     /* Sender's tag did not match the receiver's EP tag  */ 
#define EBADHANDLER       4     /* Invalid index into the recv.'s handler table      */ 
#define EBADSEGOFF        5     /* Offset into the dest-memory VM segment invalid    */
#define EBADLENGTH        6     /* Bulk xfer length goes beyond a segment's end      */
#define EBADENDPOINT      7     /* Destination endpoint does not exist               */
#define ECONGESTION       8     /* Congestion at destination endpoint                */
#define EUNREACHABLE      9     /* Destination endpoint unreachable                  */
#define EREPLYREJECTED    10    /* Destination endpoint refused reply message        */


/*
 * Op codes for the AM error handler (opcode).
 */
typedef int op_t;
#define AM_REQUEST_M      1
#define AM_REQUEST_IM     2
#define AM_REQUEST_XFER_M 3
#define AM_REPLY_M        4
#define AM_REPLY_IM       5
#define AM_REPLY_XFER_M   6

/* ------------------------------------------------------------------------------------ */
#ifdef __cplusplus
  #define BEGIN_EXTERNC extern "C" {
  #define END_EXTERNC }
#else
  #define BEGIN_EXTERNC 
  #define END_EXTERNC 
#endif

BEGIN_EXTERNC

/* AMUDP-specific user entry points */
extern int AMUDP_VerboseErrors; /* set to non-zero for verbose error reporting */
extern int AMUDP_ExpectedBandwidth; /* expected half-duplex bandwidth in KBytes/sec */
extern int AMUDP_SilentMode; /* set to non-zero to silence any non-error output */

/* set the UDP interface (local IP address) to be used for new endpoints -
 * it's necessary to call this on multi-homed hosts, otherwise endpoint creation will fail
 */
extern int AMUDP_SetUDPInterface(uint32_t IPAddress); 

/* statistical collection */
extern int AMUDP_GetEndpointStatistics(ep_t ep, amudp_stats_t *stats); /* get ep counters */
extern int AMUDP_ResetEndpointStatistics(ep_t ep); /* reset ep counters */
extern int AMUDP_AggregateStatistics(amudp_stats_t *runningsum, amudp_stats_t *newvalues); 
  /* aggregate statistics - augment running sum with the given values */
extern int AMUDP_DumpStatistics(FILE *fp, amudp_stats_t *stats, int globalAnalysis); 
  /* output stats to fp in human-readable form.
   * pass globalAnalysis non-zero if stats is a global agreggation across all nodes
   */
extern const amudp_stats_t AMUDP_initial_stats; /* the "empty" values for counters */
/* ------------------------------------------------------------------------------------ */
/* AM-2 Entry Points */

/* strictly speaking, many of these AM entry points should be true-blue functions
   (so, for example, a user could create a function pointer to them)
   but that seems a silly justification, so we went with macros on many of them 
   in the interests of performance */

/* System parameters */
#define AM_MaxShort()   AMUDP_MAX_SHORT
#define AM_MaxMedium()  AMUDP_MAX_MEDIUM
#define AM_MaxLong()    AMUDP_MAX_LONG

#define AM_MaxNumHandlers()               AMUDP_MAX_NUMHANDLERS
#define AM_MaxNumTranslations(trans)      (*(trans) = AMUDP_MAX_NUMTRANSLATIONS,AM_OK)
#define AM_MaxSegLength(nbytes)           (*(nbytes) = AMUDP_MAX_SEGLENGTH,AM_OK)

/* System initialization/termination */
extern int AM_Init();
extern int AM_Terminate();

/* endpoint/bundle management */
extern int AM_AllocateBundle(int type, eb_t *endb);
extern int AM_AllocateEndpoint(eb_t bundle, ep_t *endp, en_t *endpoint_name);
extern int AM_FreeBundle(eb_t bundle);
extern int AM_FreeEndpoint(ep_t ea);
extern int AM_MoveEndpoint(ep_t ea, eb_t from_bundle, eb_t to_bundle);

extern int AM_GetSeg(ep_t ea, void **addr, int *nbytes);
extern int AM_SetSeg(ep_t ea, void *addr, int nbytes);
extern int AM_SetTag(ep_t ea, tag_t tag);
extern int AM_GetTag(ep_t ea, tag_t *tag);

/* Translation table */
extern int AM_Map(ep_t ea, int index, en_t name, tag_t tag);
extern int AM_MapAny(ep_t ea, int *index, en_t name, tag_t tag);
extern int AM_UnMap(ep_t ea, int index);
extern int AM_GetTranslationInuse(ep_t ea, int i);
extern int AM_GetTranslationTag(ep_t ea, int i, tag_t *tag);
extern int AM_GetTranslationName(ep_t ea, int i, en_t *gan);
#define AM_GetNumTranslations(ep, pntrans)  \
  ((ep) ? ((*(pntrans) = AMUDP_MAX_NUMTRANSLATIONS), AM_OK) : AM_ERR_BAD_ARG)
#define AM_SetNumTranslations(ep, ntrans)  \
  ((ep) ? ((ntrans) == AMUDP_MAX_NUMTRANSLATIONS ? AM_OK : AM_ERR_RESOURCE) : AM_ERR_BAD_ARG)
extern int AM_SetExpectedResources(ep_t ea, int n_endpoints, int n_outstanding_requests);

/* Handler table */
extern int AM_SetHandler(ep_t ea, handler_t handler, amudp_handler_fn_t function);
extern int AM_SetHandlerAny(ep_t ea, handler_t *handler, amudp_handler_fn_t function);
#define AM_GetNumHandlers(ep, pnhandlers)  \
  ((ep) ? ((*(pnhandlers) = AMUDP_MAX_NUMHANDLERS), AM_OK) : AM_ERR_BAD_ARG) : AM_ERR_BAD_ARG)
#define AM_SetNumHandlers(ep, nhandlers)  \
  ((ep) ? ((nhandlers) == AMUDP_MAX_NUMHANDLERS ? AM_OK : AM_ERR_RESOURCE)

/* Events */
extern int AM_GetEventMask(eb_t eb, int *mask);
extern int AM_SetEventMask(eb_t eb, int mask);
extern int AM_WaitSema(eb_t eb);


/* Message interrogation */
extern int AM_GetSourceEndpoint(void *token, en_t *gan);
extern int AM_GetDestEndpoint(void *token, ep_t *endp);
extern int AM_GetMsgTag(void *token, tag_t *tagp);

/* Poll */
extern int AM_Poll(eb_t bundle);

/* Requests and Replies
   These six functions do all requests and replies.
   Macros below expand all the variants */

extern int AMUDP_Request(ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                         int numargs, ...);
extern int AMUDP_RequestI (ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                          void *source_addr, int nbytes,
                          int numargs, ...);
extern int AMUDP_RequestXfer(ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                          void *source_addr, int nbytes, int dest_offset, 
                          int async,
                          int numargs, ...);

extern int AMUDP_Reply(void *token, handler_t handler, 
                         int numargs, ...);
extern int AMUDP_ReplyI(void *token, handler_t handler, 
                          void *source_addr, int nbytes,
                          int numargs, ...);
extern int AMUDP_ReplyXfer(void *token, handler_t handler, 
                          void *source_addr, int nbytes, int dest_offset, 
                          int numargs, ...);

#define AM_Request0(ep, destep, hnum) \
   AMUDP_Request(ep, destep, hnum, 0)
#define AM_Request1(ep, destep, hnum, a0) \
   AMUDP_Request(ep, destep, hnum, 1, (int32_t)a0)
#define AM_Request2(ep, destep, hnum, a0, a1) \
   AMUDP_Request(ep, destep, hnum, 2, (int32_t)a0, (int32_t)a1)
#define AM_Request3(ep, destep, hnum, a0, a1, a2) \
   AMUDP_Request(ep, destep, hnum, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_Request4(ep, destep, hnum, a0, a1, a2, a3) \
   AMUDP_Request(ep, destep, hnum, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_Request5(ep, destep, hnum, a0, a1, a2, a3, a4) \
   AMUDP_Request(ep, destep, hnum, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_Request6(ep, destep, hnum, a0, a1, a2, a3, a4, a5) \
   AMUDP_Request(ep, destep, hnum, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_Request7(ep, destep, hnum, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_Request(ep, destep, hnum, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_Request8(ep, destep, hnum, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_Request(ep, destep, hnum, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)

#define AM_RequestI0(ep, destep, hnum, sa, cnt) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 0)
#define AM_RequestI1(ep, destep, hnum, sa, cnt, a0) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 1, (int32_t)a0)
#define AM_RequestI2(ep, destep, hnum, sa, cnt, a0, a1) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 2, (int32_t)a0, (int32_t)a1)
#define AM_RequestI3(ep, destep, hnum, sa, cnt, a0, a1, a2) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_RequestI4(ep, destep, hnum, sa, cnt, a0, a1, a2, a3) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_RequestI5(ep, destep, hnum, sa, cnt, a0, a1, a2, a3, a4) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_RequestI6(ep, destep, hnum, sa, cnt, a0, a1, a2, a3, a4, a5) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_RequestI7(ep, destep, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_RequestI8(ep, destep, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_RequestI(ep, destep, hnum, sa, cnt, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)

#define AM_RequestXfer0(ep, destep, desto, hnum, sa, cnt) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 0)
#define AM_RequestXfer1(ep, destep, desto, hnum, sa, cnt, a0) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 1, (int32_t)a0)
#define AM_RequestXfer2(ep, destep, desto, hnum, sa, cnt, a0, a1) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 2, (int32_t)a0, (int32_t)a1)
#define AM_RequestXfer3(ep, destep, desto, hnum, sa, cnt, a0, a1, a2) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_RequestXfer4(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_RequestXfer5(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_RequestXfer6(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_RequestXfer7(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_RequestXfer8(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 0, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)

#define AM_RequestXferAsync0(ep, destep, desto, hnum, sa, cnt) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 0)
#define AM_RequestXferAsync1(ep, destep, desto, hnum, sa, cnt, a0) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 1, (int32_t)a0)
#define AM_RequestXferAsync2(ep, destep, desto, hnum, sa, cnt, a0, a1) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 2, (int32_t)a0, (int32_t)a1)
#define AM_RequestXferAsync3(ep, destep, desto, hnum, sa, cnt, a0, a1, a2) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_RequestXferAsync4(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_RequestXferAsync5(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_RequestXferAsync6(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_RequestXferAsync7(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_RequestXferAsync8(ep, destep, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_RequestXfer(ep, destep, hnum, sa, cnt, desto, 1, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)

#define AM_Reply0(token, hnum) \
   AMUDP_Reply(token, hnum, 0)
#define AM_Reply1(token, hnum, a0) \
   AMUDP_Reply(token, hnum, 1, (int32_t)a0)
#define AM_Reply2(token, hnum, a0, a1) \
   AMUDP_Reply(token, hnum, 2, (int32_t)a0, (int32_t)a1)
#define AM_Reply3(token, hnum, a0, a1, a2) \
   AMUDP_Reply(token, hnum, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_Reply4(token, hnum, a0, a1, a2, a3) \
   AMUDP_Reply(token, hnum, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_Reply5(token, hnum, a0, a1, a2, a3, a4) \
   AMUDP_Reply(token, hnum, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_Reply6(token, hnum, a0, a1, a2, a3, a4, a5) \
   AMUDP_Reply(token, hnum, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_Reply7(token, hnum, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_Reply(token, hnum, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_Reply8(token, hnum, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_Reply(token, hnum, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)

#define AM_ReplyI0(token, hnum, sa, cnt) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 0)
#define AM_ReplyI1(token, hnum, sa, cnt, a0) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 1, (int32_t)a0)
#define AM_ReplyI2(token, hnum, sa, cnt, a0, a1) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 2, (int32_t)a0, (int32_t)a1)
#define AM_ReplyI3(token, hnum, sa, cnt, a0, a1, a2) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_ReplyI4(token, hnum, sa, cnt, a0, a1, a2, a3) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_ReplyI5(token, hnum, sa, cnt, a0, a1, a2, a3, a4) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_ReplyI6(token, hnum, sa, cnt, a0, a1, a2, a3, a4, a5) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_ReplyI7(token, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_ReplyI8(token, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_ReplyI(token, hnum, sa, cnt, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)

#define AM_ReplyXfer0(token, desto, hnum, sa, cnt) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 0)
#define AM_ReplyXfer1(token, desto, hnum, sa, cnt, a0) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 1, (int32_t)a0)
#define AM_ReplyXfer2(token, desto, hnum, sa, cnt, a0, a1) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 2, (int32_t)a0, (int32_t)a1)
#define AM_ReplyXfer3(token, desto, hnum, sa, cnt, a0, a1, a2) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 3, (int32_t)a0, (int32_t)a1, (int32_t)a2)
#define AM_ReplyXfer4(token, desto, hnum, sa, cnt, a0, a1, a2, a3) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 4, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3)
#define AM_ReplyXfer5(token, desto, hnum, sa, cnt, a0, a1, a2, a3, a4) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 5, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4)
#define AM_ReplyXfer6(token, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 6, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5)
#define AM_ReplyXfer7(token, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 7, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6)
#define AM_ReplyXfer8(token, desto, hnum, sa, cnt, a0, a1, a2, a3, a4, a5, a6, a7) \
   AMUDP_ReplyXfer(token, hnum, sa, cnt, desto, 8, (int32_t)a0, (int32_t)a1, (int32_t)a2, (int32_t)a3, (int32_t)a4, (int32_t)a5, (int32_t)a6, (int32_t)a7)


END_EXTERNC

#endif
