/*
 * Copyright (c) 1998 The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:              Philip Buonadonna
 * Filename:            rexec_utils.c
 * Description:		Utilities for the rexec tools...
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <ctype.h>

/* get the sysconf constant to retrieve number of processors */
#if defined(__sgi)
#define _SC_NPROCESSORS_ONLN _SC_NPROC_ONLN
#elif defined(_CRAY)
#define _SC_NPROCESSORS_ONLN _SC_CRAY_MAXPES
#elif defined(__APPLE__)
#include <sys/param.h>
#include <sys/sysctl.h>
#endif

/* get MAXHOSTNAMELEN */
#if defined(SOLARIS)
#include <netdb.h>
#else 
#include <sys/param.h>
#endif /* SOLARIS */

#ifdef REXEC
#include <netdb.h> /* gethostbyname() */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#include <string.h>
#include <tic.h>
#include <gc.h>

static int init = 0;
static int uniqueID = 0;
julong gpid = 0;

static boxInfo *boxes = NULL;
static int numBoxes = 0;
static boxInfo *myBox = NULL; 

static processInfo *processes = NULL;
static int numProcesses = 0;

extern int usingAMUDPSPMD;

/* This returns how much memory is needed for the box array */
int COMM_Init_One() {

#if defined(COMM_AMLAPI)
  		/* Init the LAPI system and get some process-local values. */
        AMLAPI_LAPI_Init ();
	do {
  	lapi_handle_t *lapi_hndl = AMLAPI_GetLAPIHandle ();	/* Get My LAPI endpoint */
  	LAPI_Qenv(*lapi_hndl, NUM_TASKS, &numBoxes);

  	if (!numBoxes) { /* Qenv(NUM_TASKS) acts funny. */
     		numBoxes = 1;
  	}  
  		/* Turn off error checking for performance. */
  	LAPI_Senv(*lapi_hndl, ERROR_CHK, 0);
  		/* Use polling instead of interrupts. */
  	LAPI_Senv(*lapi_hndl, INTERRUPT_SET, 0);
	} while (FALSE);

#else /* !COMM_AMLAPI */

  if (usingAMUDPSPMD) {
#if defined(COMM_AMUDP) 
    numBoxes = AMUDP_SPMDNumProcs();
#else
    abort();
#endif
  }
  else {
    char *variable = NULL;
  #if defined(REXEC)
    variable = getenv("REXEC_PAR_DEGREE");
    if (variable == NULL) {
      printf("Environment variable REXEC_PAR_DEGREE not defined.\nThis program must be run with rexec and the Titanium run script!\n");
      exit(1);
    }
    numBoxes = atol(variable);
  #elif defined(GLUNIX)
    numBoxes = Glib_GetParallelDegree();
    if (numBoxes < 0) {
      perror("Glib_GetParallelDegree");
      exit(1);
    }
  #else 
    numBoxes = 1;
  #endif
  }

#endif /* COMM_AMLAPI */

  init = 1;
  return numBoxes * sizeof(boxInfo);
}

/* copy the hostname and shorten it, if appropriate */
static void shortenHostname(const char *hostname, char *shorthostname) {
  strcpy(shorthostname, hostname);
  if (strchr(shorthostname,'.') && !isdigit(shorthostname[strlen(shorthostname)-1]))
    *strchr(shorthostname,'.') = '\0';
}

/* This returns how much memory is needed for the processor array */
int COMM_Init_Two(char *memory) {
	
  int i = 0, j;
  char hostname[MAXHOSTNAMELEN];
  char shorthostname[MAXHOSTNAMELEN];
  struct hostent *he = NULL;
  char *processesPerProcessors;

  boxes = (boxInfo *)memory;

#if defined(REXEC)
  if (!usingAMUDPSPMD) {
    char *envhosts = getenv("REXEC_SVRS");
    if (envhosts == NULL) {
      printf("Environment variable REXEC_SVRS not defined.\nThis program must be run with rexec and the Titanium run script!\n");
      exit(1);
    }
    { /* make a copy so we can strtok it */
      char *temp = (char *)ti_malloc(strlen(envhosts)+2);
      strcpy(temp, envhosts);
      envhosts = temp;
      }
  boxes[0].DNSname = strtok(envhosts, " \t\n"); 
  for(i = 1; i < numBoxes && (boxes[i].DNSname = strtok(NULL, " \t\n")); i++) ;
  }
#endif

#if defined(PTHREAD) || defined(THREAD)
  processesPerProcessors = (char *)getenvMaster("TI_PFORP");
  if (processesPerProcessors == NULL) {
    printf("Environment variable TI_PFORP not defined.\nThis program must be run with the Titanium run script!\n");
    exit(1);
  }
  { /* make a copy so we can strtok it */
    char *temp = (char *)ti_malloc(strlen(processesPerProcessors)+2);
    strcpy(temp, processesPerProcessors);
    processesPerProcessors = temp;
    }
#endif /* REXEC */

  for (i = 0; i < numBoxes; i++) {
    boxes[i].boxNumber = i;
    boxes[i].processesPerProcessors = 0;
  }
  
  /* Get each processes/processor tag and associate it with each box */
#if defined(PTHREAD) || defined(THREAD)
  boxes[0].processesPerProcessors = strtok(processesPerProcessors, " \t\n"); 
  for(i=1; i < numBoxes && (boxes[i].processesPerProcessors = strtok(NULL, " \t\n")); i++);
  { int gotsome = 0; /* check for missing ones */
    for(i=0; i < numBoxes; i++) {
      if (boxes[i].processesPerProcessors == NULL) {
        char *newstring = ti_malloc(4);
        strcpy(newstring, "1/1");
        boxes[i].processesPerProcessors = newstring;
        gotsome = 1;
      }
    }
    if (gotsome) {
      fprintf(stderr, "Tic: TI_PFORP too short - assuming 1/1 for unlisted processors\n");
      fflush(stderr);
    }
  }
#else
  /* no threads, but may still have multiple boxes */
  for (i = 0; i < numBoxes; i++) {
    char *newstring = ti_malloc(4);
    strcpy(newstring, "1/1");
    boxes[i].processesPerProcessors = newstring;
  }
#endif

  gethostname(hostname, MAXHOSTNAMELEN);
  shortenHostname(hostname, shorthostname);

#if defined(REXEC)  
  if (!usingAMUDPSPMD) he = gethostbyname(hostname);
#endif

  for(i = 0; i < numBoxes; i++) {
    boxInfo *box = &boxes[i];
    int thisIsMyBox = 0;

#if defined(COMM_AMUDP)
    if (usingAMUDPSPMD && AMUDP_SPMDMyProc() == i) thisIsMyBox = 1;
#endif

    if (!usingAMUDPSPMD) {
#if defined(REXEC)
      /* rexec could potentially put any hostname or IP address in REXEC_SVRS 
       * check everything that could possibly specify this box
       */
      #if 0 
        printf("box->DNSname=%s\n",box->DNSname);
        printf("trying: %s, %s\n", shorthostname, hostname);
      #endif
      thisIsMyBox = !strcasecmp(box->DNSname, shorthostname) || !strcasecmp(box->DNSname, hostname);
      if (!thisIsMyBox) {
        char temp[MAXHOSTNAMELEN];
        int i;
	shortenHostname(he->h_name, temp);
        thisIsMyBox |= !strcasecmp(box->DNSname, he->h_name) || !strcasecmp(box->DNSname, temp);
        for (i=0; he->h_aliases[i]; i++) {
          shortenHostname(he->h_aliases[i], temp);
          #if 0
            printf("trying: %s, %s\n", he->h_aliases[i], temp);
          #endif
          thisIsMyBox |= !strcasecmp(box->DNSname, he->h_aliases[i]) || !strcasecmp(box->DNSname, temp);
        }
        for (i=0; he->h_addr_list[i]; i++) {
          strcpy(temp, inet_ntoa(*(struct in_addr*)(he->h_addr_list[i])));
          #if 0
            printf("trying: %s\n", temp);
          #endif
          thisIsMyBox |= !strcmp(box->DNSname, temp);
        }
      }
#elif defined(GLUNIX)
      thisIsMyBox = (Glib_GetMyVnn() == i);
#elif defined(COMM_AMLAPI)
      int my_task_id;	
      lapi_handle_t *lapi_hndl = AMLAPI_GetLAPIHandle ();	/* Get My LAPI endpoint */
      LAPI_Qenv(*lapi_hndl, TASK_ID, &my_task_id);
      if (my_task_id == i) thisIsMyBox = 1;
#else
      assert(numBoxes == 1);
      thisIsMyBox = 1;
#endif
    }

    /* Set settings only for my box */
    if (thisIsMyBox) {
      box->pid = getpid();

#if defined(__APPLE__)
      {
        int mib[2];
        size_t len;

        mib[0] = CTL_HW;
        mib[1] = HW_NCPU;
        len = sizeof(box->totalProcessors);
        sysctl(mib, 2, &(box->totalProcessors), &len, NULL, 0);
      }
#else
      box->totalProcessors = sysconf(_SC_NPROCESSORS_ONLN);
      if (box->totalProcessors < 1) box->totalProcessors = 1; /* catch failures on Solaris */
#endif
      myBox = box;
    }

    { char *p;
      char tmp[10];
      strcpy(tmp,box->processesPerProcessors); 
      p = strchr(tmp, '/');
      if (!p) {
        fprintf(stderr, "Tic: Error - malformed TI_PFORP environment variable\n");
        exit(1);
        }
      *p = '\0';
      box->countProcesses = atol(tmp);
      numProcesses += box->countProcesses;
      box->countProcessors = atol(p+1);
    }

    /* Check for too many processors */
    if (box->totalProcessors < box->countProcessors) {
      box->countProcessors = box->totalProcessors;
    }

    /* set politeness (only polite if we're low on physical CPU's) */
    if (box->countProcesses <= box->countProcessors) politep = 0;
  }

  if (myBox == NULL) {
    printf("Failed to determine myBox number during startup.\nThis program must be run with the Titanium run script!\n");
    exit(1);
    }

  init = 2;
  return numProcesses * sizeof(processInfo);
}

void COMM_Init_Three(char *memory) {
  int i = 0, j = 0;

  processes = (processInfo *)memory;
  
  for(i = 0; i < numBoxes; i++) {
    boxInfo *box = &boxes[i];
    box->processes = (processInfo *)memory;
    memory += box->countProcesses * sizeof(processInfo);
    for(j = 0; j < box->countProcesses; j++) {
      box->processes[j].processNumber = uniqueID++;
      box->processes[j].boxProcessNumber = j;
      box->processes[j].box = box;
    }
  }
  init = 3;
}
  
void COMM_Init_Four(int argc, char **argv) {

if (!usingAMUDPSPMD) {
  #if defined(REXEC)
    char *pLLGuid = getenv("REXEC_GPID");

    if (pLLGuid == NULL) {
      printf("Environment variable REXEC_GPID not defined.\nThis program must be run with rexec and the Titanium run script!\n");
      exit(1);
    }

    for(; *pLLGuid != 0; pLLGuid++) {
      gpid = gpid * 10 + (*pLLGuid - '0');
    }
  #elif defined(GLUNIX)
    gpid = Glib_GetMyNpid();
  #else 
    gpid = getpid();
  #endif
  }

  myBox->argc = argc;
  myBox->argv = argv;

  init = 4;
}

processInfo *
COMM_GetHisBoxProcess(Process p) {
  return &myBox->processes[p];
}

processInfo *
COMM_GetHisProcess(Process p) {

  return &processes[p];
}

processInfo *
COMM_GetMyProcess() {
  return myProcess();
}

juint
COMM_GetMyGuid()
{
	return gpid & 0xFFFFFFFF;
}

julong
COMM_GetMyGuid64() 
{ 
	return gpid;
}

Box COMM_GetMyBoxNumber()
{
  return myBox->boxNumber;
}

Box
COMM_GetBoxNumberForProcNumber(Process processNumber)
{
  return processes[processNumber].box->boxNumber;
}


Box
COMM_GetHisBoxProcNumber(Process processNumber) 
{
  return processes[processNumber].boxProcessNumber;
}

Box
COMM_GetMyBParallelDegree()
{
  return myBox->countProcesses;
}

float 
COMM_GetMyProcsPerProc() 
{
  return (float)(myBox->countProcesses) / (float)(myBox->countProcessors);
}

Box
COMM_GetBParallelDegree()
{
  return numBoxes;
}

Process
COMM_GetParallelDegree() 
{
  return numProcesses;
}


Process
COMM_GetProxyProcNumberForBoxNumber(Box boxNumber)
{
  return boxes[boxNumber].processes[0].processNumber;
}

#ifndef COMM_AM2

void COMM_thread_cons_monitor(jGPointer monitor) 
{
  processInfo *pi = myProcess();
  process_monitor_table *new = (process_monitor_table *)ti_malloc(sizeof(process_monitor_table));
  
  new->monitor = monitor;
  new->nesting = 1;
  new->cdr = pi->locked_monitors;
  pi->locked_monitors = new;
}

void COMM_thread_remove_monitor(jGPointer monitor) 
{
  processInfo *pi = myProcess();
  process_monitor_table *prev, *current;

  prev = pi->locked_monitors;
  if (prev == NULL) return;
  if (___eq_ptr_check(prev->monitor, monitor)) {
    pi->locked_monitors = prev->cdr; /* don't drop the end of the list! */
    return;
  }

  current = prev->cdr;
  while(current != NULL) {
    if (___eq_ptr_check(current->monitor, monitor)) {
      prev->cdr = current->cdr;
      return;
    }
    prev = current;
    current = current->cdr;
  }
}
  
#endif /* COMM_AM2 */

