#ifndef __monitor_c
#define __monitor_c

#include <tic.h>
#include <sys/time.h>
#include <unistd.h>
#include <gc.h>
#include <native-utils.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef TRUE
	#define TRUE 1
	#define FALSE 0
#endif

#define DEBUG_MONITORS 0

#ifdef DEBUG_MONITORS
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif

static volatile int running_local = FALSE;

	/* Get the cdr proc trying for the monitor and set as owner */
int signal_try_queue( tic_monitor_t *monitor ) {
	volatile proc_queue *queue = monitor->tryingProcs;

		/* Do we have no one trying */
	if (queue == NULL) {
		#ifdef MY_DEBUG
		fprintf(stderr, "SIGNAL_TRY_QUEUE: No one to signal\n");
		#endif
		monitor->box   = (Box) -1;
		monitor->proc  = (Process) -1;
		return 0;
	}

	#ifdef MY_DEBUG
	fprintf(stderr, "SIGNAL_TRY_QUEUE: signal process %d\n", queue->proc);
	#endif
	monitor->tryingProcs = queue->cdr;
        monitor->box         = queue->box;
        monitor->proc        = queue->proc;
        ti_free( (proc_queue*) queue );
        return 1;
}

void insert_try_queue( tic_monitor_t *monitor, Box box, Process proc ) {
        struct proc_queue *try_queue     = ti_malloc(sizeof(struct proc_queue));
	volatile proc_queue *tryingProcs = monitor->tryingProcs;
        try_queue->proc = proc;
        try_queue->box  = box;
        try_queue->cdr  = NULL;

	if (tryingProcs == NULL) {
		monitor->tryingProcs = try_queue;
	} else {
		for ( ;tryingProcs->cdr != NULL; tryingProcs = tryingProcs->cdr) 
			;
		tryingProcs->cdr = try_queue;
	}
}

	/* Move one proc to trying list */
void signal_wait_queue( tic_monitor_t *monitor ) {
        Process proc; 
	Box box;
	volatile proc_queue *queue = monitor->waitingProcs;

		/* Do we have someone waiting */
	if (queue == NULL) {
  		#ifdef WAIT_DEBUG
		fprintf(stderr, "\t\tSIGNAL_WAIT_QUEUE: no one waiting\n");
		#endif
		return;
	}

		/* Get cdr proc and move to the end of the trying list */
  	#ifdef WAIT_DEBUG
        fprintf(stderr, "\t\tSINGAL_WAIT_QUEUE: Process %d signaled\n", queue->proc);
	#endif
	proc = queue->proc;
	box  = queue->box;
	monitor->waitingProcs = queue->cdr; 	/* Remove proc from the waiting list */
	insert_try_queue( monitor, box, proc );	/* Add proc to the trying list */
	ti_free( (proc_queue*) queue );
}

int cancel_waiting_proc( tic_monitor_t *monitor, Process proc ) {
	volatile proc_queue *current = monitor->waitingProcs;
	volatile proc_queue *prev    = current;

		/* Do we have someone waiting */
	if (current == NULL) {
		return FALSE;
	}
	if (current->proc == proc) { 	/* proc is first in the wait list */
		monitor->waitingProcs = current->cdr;
		ti_free( (proc_queue*) current );
		return TRUE;
	} else {			/* search for proc and remove from wait list */
		for(current = current->cdr; current != NULL; current = current->cdr) {
			if(current->proc == proc) {
				prev->cdr = current->cdr;
				ti_free((proc_queue*)  current );
				return TRUE;
			}
			prev = current;
		}
	}
	return FALSE;
}

void insert_wait_queue( tic_monitor_t *monitor, Box box, Process proc ) {
	volatile proc_queue *waitingProcs = monitor->waitingProcs;
        proc_queue *wait_queue            = ti_malloc(sizeof(proc_queue));
        wait_queue->proc = proc;
        wait_queue->box  = box;
        wait_queue->cdr  = NULL;

	if (waitingProcs == NULL) {
		monitor->waitingProcs = wait_queue;
	} else {	/* Insert at the end of the list */
		for ( ;waitingProcs->cdr != NULL; waitingProcs = waitingProcs->cdr) 
			;
		waitingProcs->cdr = wait_queue;
	}
}

	/* Move waiting procs to try list */
void broadcast_wait_queue( tic_monitor_t *monitor ) {
	volatile proc_queue *trying_queue  = monitor->tryingProcs;
	volatile proc_queue *waiting_queue = monitor->waitingProcs;
	
	if (trying_queue == NULL) {
		monitor->tryingProcs = waiting_queue;
	} else {
		for ( ;trying_queue->cdr != NULL; trying_queue = trying_queue->cdr) 
			;
		trying_queue->cdr = waiting_queue;
	}
	monitor->waitingProcs = NULL;
}

	/* Get process' PMT for monitor */
struct process_monitor_table* get_pmt(jGPointer monitor) {
      struct process_monitor_table *my_pmt = COMM_GetMyProcess()->locked_monitors;
      for( ; my_pmt != NULL; my_pmt = my_pmt->cdr)
	if (EQUAL_GLOBAL(my_pmt->monitor, monitor)) break;
      return my_pmt;
}

void insert_monitor_pmt(jGPointer monitor) {
      process_monitor_table *pmt = ti_malloc(sizeof(process_monitor_table));
      pmt->monitor = monitor;
      pmt->nesting = 1;
      pmt->cdr     = COMM_GetMyProcess()->locked_monitors;
      COMM_GetMyProcess()->locked_monitors = pmt;
}

void remove_monitor_from_pmt(jGPointer monitor) {
      process_monitor_table *current = COMM_GetMyProcess()->locked_monitors;
      if (current == NULL) {	/* Do we own any locks? */
	return;
      } else {			/* Search for the lock */
	if (EQUAL_GLOBAL(current->monitor, monitor)) {
		COMM_GetMyProcess()->locked_monitors = current->cdr;
 		ti_free( (process_monitor_table*) current );
      	} else {
        	process_monitor_table *prev = current;
		for(current = current->cdr; current != NULL; current = current->cdr) {
      	   		if (EQUAL_GLOBAL(current->monitor, monitor)) {
				prev->cdr = current->cdr;	
				ti_free((process_monitor_table*) current);
				return;
           		}
	   		prev = current;
		}
	}
      }
}

void monitor_startup() {
	COMM_GetMyProcess()->locked_monitors      = NULL;	
	COMM_GetMyProcess()->signal_proc          = (Process)-1;	
	COMM_GetMyProcess()->signal_box           = (Process)-1;	
	COMM_GetMyProcess()->waiting_for_monitor  = 0;	
}

void monitor_destroy(tic_monitor_t *monitor) {

  volatile proc_queue *temp;

  if (monitor != NULL) {
    /* Hopefully, the monitor will never be deallocated before */
    /* replies to all processes waiting to enter the monitor are sent. */
    /* However, this is done just in case this situation arises. */
    while (monitor->waitingProcs != NULL) {
      temp = monitor->waitingProcs;
      monitor->waitingProcs = monitor->waitingProcs->cdr;
      temp->cdr = NULL;
      ti_free((proc_queue*)temp);
    }
    while (monitor->tryingProcs != NULL) {
      temp = monitor->tryingProcs;
      monitor->tryingProcs = monitor->tryingProcs->cdr;
      temp->cdr = NULL;
      ti_free((proc_queue*)temp);
    }
  }

}


void monitor_enter(jGPointer monitor) {

  struct process_monitor_table *my_pmt = get_pmt(monitor);
  tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);
  Box box = TO_BOX(monitor);

    /* Do I already own it? */
  if (my_pmt != NULL) {
      my_pmt->nesting++;
      return;
  }

  COMM_GetMyProcess()->waiting_for_monitor = WAIT;
  if (box == MYBOX) {	/* Call monitor_enter handler directly */
    __tic_lock(&AM_lock);
    running_local = TRUE;
    __monitor_enter_request_handler(NULL, mon, MYBOX, MYPROC);
    running_local = FALSE;
    __tic_unlock(&AM_lock);
  } else { 		/* The monitor is on another box */
      TI_AM_Request3(CUR_ENDPOINT, box, __MON_ENTER_H, (jUIntPointer)mon, MYBOX, MYPROC);
  }
  while ( COMM_GetMyProcess()->waiting_for_monitor != GRANTED )
	tic_poll();
  insert_monitor_pmt(monitor);
}


___MON_HANDLER2(enter_request)

void __monitor_enter_request_handler(void *token, tic_monitor_t *mon, 
				     Box box, Process proc) {
   if ( mon->proc == (Process)-1 ) {
      mon->proc = proc;
      mon->box  = box;

	/* Send reply, which will set waiting_for_monitor = GRANTED */
      if (running_local) {	/* Call reply handler directly */
	__monitor_enter_reply_handler(NULL, GRANTED, proc);
      } else {			/* Send reply via AM */
      	TI_AM_Reply2(token, __MON_ENTER_RH, GRANTED, proc);
      }
    } else {	/* Process must wait for the monitor */
      insert_try_queue(mon, box, proc); 
      if (running_local) {	/* Call reply handler directly */
	__monitor_enter_reply_handler(NULL, WAIT, proc);
      } else {			/* Send reply via AM */
      	TI_AM_Reply2(token, __MON_ENTER_RH, WAIT, proc);
      }
    }
}

___MON_REPLY_HANDLER2(enter_reply)

	/* Signal process only if granted (process knows if it is waiting) */
void __monitor_enter_reply_handler(void *token, int value, Process proc) {
  if (value != WAIT) {
  	ASSERT( COMM_GetHisProcess(proc) != NULL );
  	COMM_GetHisProcess(proc)->waiting_for_monitor = value;
  }
}

void monitor_exit(jGPointer monitor) {

  process_monitor_table* pmt = get_pmt(monitor);

  if (pmt == NULL) { 
  	fprintf(stderr, "Tried to exit on a lock you didn't own.\n");
  	tossMonitorStateException_str("Tried to exit on a lock you didn't own.\n");
  }

  if (pmt->nesting > 1) {
 	pmt->nesting--;
  } else {	/* Release monitor */
  	tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);
  	Box box = TO_BOX(monitor);
	remove_monitor_from_pmt(monitor);

	COMM_GetMyProcess()->signal_proc = -2;
	COMM_GetMyProcess()->signal_box  = -2;
  	#ifdef MY_DEBUG
	fprintf(stdout, "\t\tMONITOR_EXIT: calling exit monitor request\n");
	#endif
  	if (box == MYBOX) {
    		__tic_lock(&AM_lock);
    		running_local = TRUE;
    		__monitor_exit_request_handler(NULL, mon, MYBOX, MYPROC);
    		running_local = FALSE;
    		__tic_unlock(&AM_lock);
  	} else {
    		TI_AM_Request3(CUR_ENDPOINT, box, __MON_EXIT_H, 
		   	(jUIntPointer)mon, MYBOX, MYPROC);
	}
  	#ifdef MY_DEBUG
	fprintf(stdout, "\t\tMONITOR_EXIT: wait for next process to signal\n");
	#endif

    	while (COMM_GetMyProcess()->signal_proc == (Process)-2)
      		tic_poll();

	if (COMM_GetMyProcess()->signal_proc != (Process)-1) { /* Signal next proc */
  		#ifdef MY_DEBUG
		fprintf(stdout, "\t\tMONITOR_EXIT: Process %d signal process %d has the monitor\n", 
					MYPROC, COMM_GetMyProcess()->signal_proc);
		#endif
  		if (COMM_GetMyProcess()->signal_box == MYBOX) {
    			__tic_lock(&AM_lock);
    			running_local = TRUE;
    			__monitor_signal_request_handler(NULL, COMM_GetMyProcess()->signal_proc);
    			running_local = FALSE;
    			__tic_unlock(&AM_lock);
		} else {
      			TI_AM_Request1(CUR_ENDPOINT, COMM_GetMyProcess()->signal_box, 
			       __MON_SIGNAL_H, COMM_GetMyProcess()->signal_proc);
		}
	} else {
  		#ifdef MY_DEBUG
		fprintf(stdout, "\t\tMONITOR_EXIT: No next process to signal\n");
		#endif
	}
  }
}

___MON_HANDLER3(signal_request)

void __monitor_signal_request_handler(void *token, Process proc) {
    COMM_GetHisProcess(proc)->waiting_for_monitor = GRANTED;

    #ifdef WAIT_DEBUG
    fprintf(stderr, "SIGNAL_REQUEST_HANDLER: Signal process %d has the lock\n", proc); 
    #endif

    if (running_local) {   /* Reply to the signaller */
	__monitor_signal_reply_handler(NULL, proc);
    } else {
    	TI_AM_Reply1(token, __MON_SIGNAL_RH, proc);
    }
}

___MON_REPLY_HANDLER3(signal_reply)

void __monitor_signal_reply_handler(void *token, Process proc) {
	/* Do nothing */
}

___MON_HANDLER2(exit_request)

void __monitor_exit_request_handler(void *token, tic_monitor_t *mon,
				    Box box, Process proc) {
  	/* Does he own it? */
  if (mon->box == box && mon->proc == proc) {
    signal_try_queue( mon );
    if (running_local) {   /* Reply with the next process to get the monitor */
	__monitor_exit_reply_handler(NULL, proc, mon->proc, mon->box);
    } else {
    	TI_AM_Reply3(token, __MON_EXIT_RH, proc, mon->proc, mon->box);
    }
  } else { /* Can't happen. */
	fprintf(stderr, "Attempt to exit a monitor that process does not own\n");
	ASSERT( FALSE );
  }
}

___MON_REPLY_HANDLER1(exit_reply)

void __monitor_exit_reply_handler(void *token, Process proc,
				  Process signal_proc, Box signal_box) {
  ASSERT( COMM_GetHisProcess(proc) != NULL );
  #ifdef MY_DEBUG
  fprintf(stdout, "\tEXIT_REPLY_HANDLER: lock released signal process %d\n", (int)signal_proc);
  #endif
  COMM_GetHisProcess(proc)->signal_box  = signal_box;
  COMM_GetHisProcess(proc)->signal_proc = signal_proc;
}

/***************************************************************************/

void monitor_wait(jGPointer monitor, ___tic_time_t *tm) {
  Box box; int my_nesting; tic_monitor_t *mon;
  process_monitor_table* pmt;

  #ifdef WAIT_DEBUG
  fprintf(stderr, "monitor_wait: get_pmt\n");
  #endif
  pmt = get_pmt(monitor);

  if (pmt == NULL) 
  	tossMonitorStateException_str("Tried to wait on a lock you didn't own.\n");

  #ifdef WAIT_DEBUG
  fprintf(stderr, "Monitor wait called for process %d nesting %d\n", MYPROC, pmt->nesting);
  #endif

  my_nesting = pmt->nesting;			/* Save my nesting */

  #ifdef WAIT_DEBUG
  fprintf(stderr, "MONITOR_WAIT: Process %d has nesting %d on entry\n", MYPROC, my_nesting); 
  #endif

  remove_monitor_from_pmt(monitor);		/* Remove monitor from list of owned monitors */
  mon = (tic_monitor_t *)TO_LOCAL(monitor);

  box = TO_BOX(monitor);

  COMM_GetMyProcess()->signal_proc = -2;		/* Will poll until this becomes > -2 */
  COMM_GetMyProcess()->signal_box  = -2;
  COMM_GetMyProcess()->waiting_for_monitor = WAIT; 	/* Will need to contend for the lock */

	/* Release monitor and get who to signal next */
  if (box == MYBOX) {
  	__tic_lock(&AM_lock);
    	running_local = TRUE;
    	__monitor_wait_request_handler(NULL, mon, MYBOX, MYPROC);
    	running_local = FALSE;
    	__tic_unlock(&AM_lock);
  } else {
  	TI_AM_Request3(CUR_ENDPOINT, box, __MON_WAIT_H, (jUIntPointer)mon, MYBOX, MYPROC);
  }

  #ifdef WAIT_DEBUG
  fprintf(stderr, "Wait request complete process %d\n", MYPROC);
  #endif

	/* Poll until I know who to signal next */
  while (COMM_GetMyProcess()->signal_proc == (Process)-2)
  	tic_poll();

  #ifdef WAIT_DEBUG
  fprintf(stderr, "Signal process %d has the lock\n", COMM_GetMyProcess()->signal_proc);
  #endif

	/* Signal next process if one exists */
  if (COMM_GetMyProcess()->signal_proc != (Process)-1) { /* Signal next proc */
  	if (COMM_GetMyProcess()->signal_box == MYBOX) {
    		__tic_lock(&AM_lock);
    		running_local = TRUE;
    		__monitor_signal_request_handler(NULL, COMM_GetMyProcess()->signal_proc);
    		running_local = FALSE;
    		__tic_unlock(&AM_lock);
	} else {
      		TI_AM_Request1(CUR_ENDPOINT, COMM_GetMyProcess()->signal_box, 
		       __MON_SIGNAL_H, COMM_GetMyProcess()->signal_proc);
	}
  }

  #ifdef WAIT_DEBUG
  fprintf(stderr, "Process %d has been signaled by %d\n", COMM_GetMyProcess()->signal_proc, MYPROC);
  #endif
	/* If we have a time to wait then poll while ___tic_millisuntil(___tic_time_t tm) > 0 */
  if (tm != NULL) {
        #ifdef TIME_WAIT_DEBUG
	fprintf(stderr, "Process %d will now wait for a given amount of time\n", MYPROC);
        #endif
		/* Wait for a given amount of time */
  	while ( COMM_GetMyProcess()->waiting_for_monitor != GRANTED && ___tic_millisuntil(tm) >= 0 )
		tic_poll();

        #ifdef TIME_WAIT_DEBUG
 	fprintf(stderr, "Process %d done waiting\n", MYPROC);
        #endif

		/* Cancel the wait for monitor, and enter monitor if not on try list */
  	if ( COMM_GetMyProcess()->waiting_for_monitor != GRANTED ) {
	   if (box == MYBOX) {
  		__tic_lock(&AM_lock);
    		running_local = TRUE;
    		__monitor_cancel_wait_request_handler(NULL, mon, MYBOX, MYPROC);
    		running_local = FALSE;
    		__tic_unlock(&AM_lock);
  	   } else {
        	TI_AM_Request3(CUR_ENDPOINT, box, __MON_CANCEL_WAIT_H, (jUIntPointer)mon, MYBOX, MYPROC);
  	   }
	}
  }
	/* Poll until I get the monitor back */
  #ifdef WAIT_DEBUG
  fprintf(stderr, "Process %d: Poll until I get the monitor back\n", MYPROC);
  #endif

  while ( COMM_GetMyProcess()->waiting_for_monitor != GRANTED )
	tic_poll();
  insert_monitor_pmt(monitor);
  pmt = get_pmt(monitor);
  ASSERT( pmt != NULL );
  pmt->nesting = my_nesting;
  #ifdef WAIT_MONITOR
  fprintf(stderr, "MONITOR_WAIT: Process %d has nesting %d on exit\n", MYPROC, my_nesting); 
  #endif
}

___MON_HANDLER2(wait_request)

void __monitor_wait_request_handler(void *token, tic_monitor_t *mon,
				    Box box, Process proc) {
  #ifdef WAIT_DEBUG
  fprintf(stderr, "WAIT_REQUEST_HANDLER: process %d\n", proc);
  #endif
  /* Does he own it? */
  if (mon->box == box && mon->proc == proc) {
    insert_wait_queue(mon, box, proc);
    signal_try_queue( mon );
    #ifdef WAIT_DEBUG
    fprintf(stderr, "WAIT_REQUEST_HANDLER: process %d signal process %d\n", proc, mon->proc);
    #endif
    if (running_local) {   /* Reply with the next process to get the monitor */
	__monitor_wait_reply_handler(NULL, proc, mon->proc, mon->box);
    } else {
    	TI_AM_Reply3(token, __MON_WAIT_RH, proc, mon->proc, mon->box);
    }
  } else {
    fprintf(stderr, "Process request to wait on a monitor, which he did not own\n");
    ASSERT( FALSE );
  }
}

___MON_REPLY_HANDLER1(wait_reply)

void __monitor_wait_reply_handler(void *token, Process proc, 
				Process signal_proc, Box signal_box) {
  #ifdef WAIT_DEBUG
  fprintf(stderr, "WAIT_REPLY_HANDLER: process %d will signal process %d\n", proc, signal_proc);
  #endif
  COMM_GetHisProcess(proc)->signal_proc = signal_proc;
  COMM_GetHisProcess(proc)->signal_box  = signal_box;
}


void monitor_notify(jGPointer monitor) {
    process_monitor_table* pmt = get_pmt(monitor);
    tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);
    Box box = TO_BOX(monitor);

    #ifdef WAIT_DEBUG
    fprintf(stderr, "\tMONITOR_NOTIFY: process %d\n", MYPROC);
    #endif

    if (pmt == NULL) 
  	tossMonitorStateException_str("Tried to notify a process on a lock you didn't own.\n");

    if (box == MYBOX) {
    	__tic_lock(&AM_lock);
    	running_local = TRUE;
    	__monitor_notify_request_handler(NULL, mon);
    	running_local = FALSE;
    	__tic_unlock(&AM_lock);
    } else {
      	TI_AM_Request1(CUR_ENDPOINT, box, __MON_NOTIFY_H, (jUIntPointer)mon); 
    }
}

___MON_HANDLER2(cancel_wait_request)

void __monitor_cancel_wait_request_handler(void *token, tic_monitor_t *mon, 
						Box box, Process proc) {
   int wait_canceled = cancel_waiting_proc(mon, proc);
   if ( wait_canceled )
      	insert_try_queue(mon, box, proc); 

   if (mon->proc == (Process)-1 && mon->tryingProcs != NULL) {
      signal_try_queue( mon );
      ASSERT( mon->proc == proc );
	/* Send reply, which will set waiting_for_monitor = GRANTED */
      if (running_local) {	/* Call reply handler directly */
	__monitor_enter_reply_handler(NULL, GRANTED, proc);
      } else {			/* Send reply via AM */
      	TI_AM_Reply2(token, __MON_ENTER_RH, GRANTED, proc);
      }
    } else {	/* Process must wait for the monitor */
      if (running_local) {	/* Call reply handler directly */
	__monitor_enter_reply_handler(NULL, WAIT, proc);
      } else {			/* Send reply via AM */
      	TI_AM_Reply2(token, __MON_ENTER_RH, WAIT, proc);
      }
    }
}

___MON_HANDLER1(notify_request)

void __monitor_notify_request_handler(void *token, tic_monitor_t *mon) {
      #ifdef WAIT_DEBUG
      fprintf(stderr, "NOTIFY_REQUEST_HANDLER: process %d owns monitor\n", mon->proc);
      #endif
      signal_wait_queue(mon);
      if (running_local) {	/* Call reply handler directly */
	__monitor_notify_reply_handler(NULL, mon->proc);
      } else {			/* Send reply via AM */
      	TI_AM_Reply1(token, __MON_NOTIFY_RH, mon->proc);
      }
}

void monitor_notify_all(jGPointer monitor) {
    process_monitor_table* pmt = get_pmt(monitor);
    tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);
    Box box = TO_BOX(monitor);
    #ifdef WAIT_DEBUG
    fprintf(stderr, "MONITOR_NOTIFY_ALL\n");
    #endif
    if (pmt == NULL) 
  	tossMonitorStateException_str("Tried to notify all processes on a lock you didn't own.\n");

    if (box == MYBOX) {
    	__tic_lock(&AM_lock);
    	running_local = TRUE;
    	__monitor_notify_all_request_handler(NULL, mon, MYBOX, MYPROC);
    	running_local = FALSE;
    	__tic_unlock(&AM_lock);
    } else {
      	TI_AM_Request3(CUR_ENDPOINT, box, __MON_NOTIFY_ALL_H, 
				(jUIntPointer)mon, MYBOX, MYPROC); 
    }
}

___MON_HANDLER2(notify_all_request)

void __monitor_notify_all_request_handler(void *token, tic_monitor_t *mon,
					  Box box, Process proc) {
      broadcast_wait_queue(mon);
      if (running_local) {	/* Call reply handler directly */
	__monitor_notify_reply_handler(NULL, proc);
      } else {			/* Send reply via AM */
      	TI_AM_Reply1(token, __MON_NOTIFY_RH, proc);
      }
}
 
___MON_REPLY_HANDLER3(notify_reply)

void __monitor_notify_reply_handler(void *token, Process proc) {
	/* Nothing to do */
}

#endif  /* __monitor_c */
