/*									tab:8
 *
 * memory.h - Global memory accesses with acks (get, put, read and write)
 *            Implementation dependent definitions and declarations.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Arvind Krishnamurthy
 * Version:		69
 * Creation Date:	Tue Sep 22 23:54:02 1992
 * Filename:		memory.h
 * History:
 *	LTL	67	Mon Jul 24 21:13:24 1995
 *		Removed #ifdef PMAM
 *	ssl	49	Mon Mar 21 09:47:12 1994
 *		sync for HPAM can use poll but PMAM cannot
 *	ssl	48	Sun Mar 20 14:22:13 1994
 *		sync on hps can use poll now
 *	ssl	46	Fri Mar 11 16:12:35 1994
 *		sync must use am_wait_until_zero for hpam, because am_poll does
 *      not retransmit
 *	ssl	45	Sun Dec  5 23:02:43 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	44	Sat Nov 27 22:57:58 1993
 *		Moved __get* __put* __write to bulk.h. Included tic.h
 *	ssl	43	Tue Nov 23 03:26:44 1993
 *		__d_get_ctr, __d_put, ... store int's instead of doubles.
 *      i860 passes doubles in fp registers.
 *	ssl	42	Tue Nov 16 12:00:34 1993
 *		Replaced global tags with macros.
 *	ssl	41	Thu Oct 28 21:59:55 1993
 *		Include am.h instead of cm5.h
 *      Took out __put_rvalue stuff.
 *	AK	40	Thu Jun 17 15:13:53 1993
 *		Commented out the check for < 8 in __get_ctr.
 *		Compiler does the check currently.
 *	SL	38	Thu May 13 17:01:47 1993
 *		Added SC_PROFILE information
 *	AK	37	Tue Mar 16 14:33:33 1993
 *		_*_struct_reads return pointers, not structures.
 *	AK	34	Sat Jan 23 01:47:53 1993
 *		Modified __read to report errors.
 * 	TvE	32	Sat Dec 19 00:15:28 PST 1992
 *		Started major clean-up and overhaul.
 *		Extracted from tic_int.h
 *	AK	1	Sat Nov 28 21:17:44 1992
 *		changed include file name from cmam2-4
 * 	TvE	31	Tue Sep 29 12:47:46 PDT 1992
 *		changed include file names
 */

#ifndef __MEMORY_H
#define __MEMORY_H

/* .....................
 *
 * Profiling information
 *
 * .....................
 */

#include <stdio.h>
#include <tic.h>



extern __tic_lock_t GetPut_mutex;

EXTERN INLINE_MODIFIER void __get_ctr(void *lPtr, jGPointer gPtr, jint len,
							 Counter *ctr);
EXTERN INLINE_MODIFIER void __put_ctr(jGPointer gPtr, void *lPtr, int len,
							 Counter *ctr);
EXTERN INLINE_MODIFIER void __put(jGPointer gPtr, void *lPtr, int len);
EXTERN INLINE_MODIFIER void __ticwrite(jGPointer gPtr, void *lPtr, int len);

/* .......................................
 *
 * Keep track of outstanding GETs and PUTs
 *
 * .......................................
 */

/* A local variable which is incremented and then we wait for it to go to zero.
   Used as the implicit flag in all non-blocking operations */
extern Counter numGetOuts;
extern Counter numPutOuts;

EXTERN INLINE_MODIFIER void __decr_ctr_reply_handler(void *token __attribute__((unused)), Counter *ctr)
INLINE({

  __tic_lock(&GetPut_mutex);
  *ctr -= 1;
  __tic_unlock(&GetPut_mutex);
})



/* Called from within reply handlers in bulk.c and memory.c */
EXTERN INLINE_MODIFIER void __decr_ctr(Counter *ctr)
INLINE({
  __tic_lock(&GetPut_mutex);
  *ctr -= 1;
  __tic_unlock(&GetPut_mutex);
})

/* Called from get/put routines in memory.h */
EXTERN INLINE_MODIFIER void __incr_ctr(Counter *ctr)
INLINE({
  __tic_lock(&GetPut_mutex);
  *ctr += 1;
  __tic_unlock(&GetPut_mutex);
})

EXTERN INLINE_MODIFIER void __add_ctr(Counter *ctr, int amt)
INLINE({
  __tic_lock(&GetPut_mutex);
  *ctr += amt;
  __tic_unlock(&GetPut_mutex);
})

extern void __decr_put_ctr_reply_handler(void *token);

/* .........................
 *
 * Synchronize GETs and PUTs
 *
 * .........................
 */


/* TODO: the two functions below need to be fixed to use 
   thread-specific counters to avoid livelock on CLUMPS */

#ifdef COMM_AM2
EXTERN INLINE_MODIFIER void tic_read_sync(void)
INLINE({
  while (numGetOuts) {
    tic_poll();
  }
}) 
#else
#define tic_read_sync()
#endif

#ifdef COMM_AM2
EXTERN INLINE_MODIFIER void tic_write_sync(void)
INLINE({
  while (numPutOuts) {
    tic_poll();
  }
}) 
#else
#define tic_write_sync()
#endif

#define tic_sync() do {	\
  tic_read_sync();   	\
  tic_write_sync(); 	\
  } while (0)

/* Check whether all of this node's split-phase ops have completed */
EXTERN INLINE_MODIFIER int is_sync(void)
INLINE({
  tic_poll();
  return !(numGetOuts || numPutOuts);
})

/* a local memory barrier - ensure all previous stores to local mem
   from this proc are globally completed across this SMP */
#ifndef SHARED_MEMORY
 #define tic_local_memory_sync()
#else
#ifdef __GNUC__
 #define ASM(mnemonic) asm volatile (#mnemonic)
#else
 #define ASM(mnemonic) asm { mnemonic }
#endif

#if defined(__sparc__) || defined(__sparc) || defined(sparc)
 EXTERN INLINE_MODIFIER void tic_local_memory_sync(void) INLINE({
  ASM(stbar); /* SPARC store barrier */
 })
#elif defined(__mips__) || defined(__mips) || defined(mips) || defined(_MIPS_ISA)
 EXTERN INLINE_MODIFIER void tic_local_memory_sync(void) INLINE({
  ASM(sync);  /* MIPS II+ memory barrier */ 
 })
#elif defined(__i386__) || defined(__i386) || defined(i386) || \
      defined(__i486__) || defined(__i486) || defined(i486) || \
      defined(__i586__) || defined(__i586) || defined(i586) || \
      defined(__i686__) || defined(__i686) || defined(i686)
 #if 0
  EXTERN INLINE_MODIFIER void tic_local_memory_sync(void) INLINE({
   ASM(mfence); /* only works on pentiums and higher? */
  })
 #else
  EXTERN INLINE_MODIFIER void tic_local_memory_sync(void) INLINE({
   ASM(sfence);
  })
 #endif
#elif defined(_POWER)
 EXTERN INLINE_MODIFIER void tic_local_memory_sync(void) INLINE({
  ASM(dcs); /* IBM SP POWER2, POWER3 */
 })
#elif defined(_POWERPC)
 EXTERN INLINE_MODIFIER void tic_local_memory_sync(void) INLINE({
  ASM(sync); /* IBM PowerPC */
 })
#else
 #error unknown CPU - dont know how to do a local memory barrier
#endif
#endif

/* ..........................
 *
 * GETs with explicit counter
 *
 * ..........................
 */

/* Handler on the remote end to send value back */
extern void __b_get_handler (void *token, jbyte *addr, jbyte *retAddr, Counter *ctr);
extern void __sh_get_handler(void *token, jshort *addr, jshort *retAddr, Counter *ctr);
extern void __i_get_handler (void *token, jint *addr, jint *retAddr, Counter *ctr);
extern void __f_get_handler (void *token, jfloat *addr, jfloat *retAddr, Counter *ctr);
extern void __d_get_handler (void *token, jdouble *addr, jdouble *retAddr, Counter *ctr);
extern void __l_get_handler(void *token, jlong *addr,
			    jlong *retAddr, Counter *ctr);
extern void __lp_get_handler(void *token, void **addr,
			     void **retAddr, Counter *ctr);
extern void __gp_get_handler(void *token, jGPointer *addr,
			     jGPointer *retAddr, Counter *ctr);


/* Handlers on the local end to receive value */
extern void __b_get_reply_handler (void *token, jbyte *addr, jbyte data,
				   Counter *ctr);
extern void __sh_get_reply_handler(void *token, jshort *addr, jshort data,
				   Counter *ctr);
extern void __i_get_reply_handler (void *token, jint *addr, jint data,
				   Counter *ctr);
extern void __f_get_reply_handler (void *token, jfloat *addr, jfloat data,
				   Counter *ctr);
extern void __d_get_reply_handler (void *token, jdouble *addr,  
				   jdouble data, Counter *ctr);
extern void __l_get_reply_handler(void *token, jlong *addr,
				  jlong data, Counter *ctr);

extern void __lp_get_reply_handler(void *token, void **addr,
				  void *data, Counter *ctr);

#ifdef PTR64
extern void __gp_get_reply_handler(void *token, jGPointer *addr,
				   jUIntPointer word0, jUIntPointer word1, 
				   Counter *ctr);
#else
extern void __gp_get_reply_handler(void *token, jGPointer *addr,
				   jGPointer data, 
				   Counter *ctr);
#endif


/* GET character with explicit counter */
EXTERN INLINE_MODIFIER void __b_get_ctr(jbyte *lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
     INLINE({
       int box = tobox(gPtr);
       jbyte *addr = tolocal(gPtr);
       
       if (box == MYBOX) {
	 *lPtr = *addr;
       } else {
	 __incr_ctr(ctr);

	 TI_AM_Request3(CUR_ENDPOINT, box, __B_GET_H, (jUIntPointer)addr, 
			(jUIntPointer)lPtr, (jUIntPointer)ctr);

       }
     })
#else
     INLINE({
       jbyte *addr = (jbyte *)tolocal(gPtr);
       *lPtr = *addr;
     })
#endif /* COMM_AM2 */

/* GET short with explicit counter */
EXTERN INLINE_MODIFIER void __sh_get_ctr(jshort *lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
  int box = tobox(gPtr);
  jshort *addr = tolocal(gPtr);
  
  if (box == MYBOX) {
    *lPtr = *addr;
  } else {
    __incr_ctr(ctr);
    
    TI_AM_Request3(CUR_ENDPOINT, box, __SH_GET_H, (jUIntPointer)addr, 
		   (jUIntPointer)lPtr, (jUIntPointer)ctr);
  }
})
#else
     INLINE({
       jshort *addr = tolocal(gPtr);
       
       *lPtr = *addr;
     })
#endif /* COMM_AM2 */


/* GET int with explicit counter */
EXTERN INLINE_MODIFIER void __i_get_ctr(jint *lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
  int box = tobox(gPtr);
  jint *addr = tolocal(gPtr);
  
  if (box == MYBOX) {
    *lPtr = *addr;
  } else {
    __incr_ctr(ctr);
    
    TI_AM_Request3(CUR_ENDPOINT, box, __I_GET_H, (jUIntPointer)addr, 
		   (jUIntPointer)lPtr, (jUIntPointer)ctr);
  }
})
#else
     INLINE({
       jint *addr = tolocal(gPtr);
       *lPtr = *addr;
     })
#endif /* COMM_AM2 */

/* GET float with explicit counter */
EXTERN INLINE_MODIFIER void __f_get_ctr(jfloat *lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
  int box = tobox(gPtr);
  jfloat *addr = tolocal(gPtr);
  
  if (box == MYBOX) {
    *lPtr = *addr;
  } else {
    __incr_ctr(ctr);
    
    TI_AM_Request3(CUR_ENDPOINT, box, __F_GET_H, (jUIntPointer)addr, 
		(jUIntPointer)lPtr, (jUIntPointer)ctr);
  }
})
#else
INLINE({
  jfloat *addr = tolocal(gPtr);
  
    *lPtr = *addr;
})
#endif /* COMM_AM2 */

/* GET float with explicit counter */
EXTERN INLINE_MODIFIER void __d_get_ctr(jdouble *lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
	int box = tobox(gPtr);
	jdouble *addr = tolocal(gPtr);

	if (box == MYBOX) {
	  *lPtr = *addr;
	} else {
	  __incr_ctr(ctr);
		
		TI_AM_Request3(CUR_ENDPOINT, box, __D_GET_H, (jUIntPointer)addr, 
			       (jUIntPointer)lPtr, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jdouble *addr = tolocal(gPtr);
	  *lPtr = *addr;
})
#endif /* COMM_AM2 */


/* GET long with explicit counter */
EXTERN INLINE_MODIFIER void __l_get_ctr(jlong *lPtr, jGPointer gPtr,Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
	int box = tobox(gPtr);
	jlong *addr = tolocal(gPtr);

	if (box == MYBOX) {
	  *lPtr = *addr;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __L_GET_H, (jUIntPointer)addr, 
		      (jUIntPointer)lPtr, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jlong *addr = tolocal(gPtr);

	  *lPtr = *addr;
})
#endif /* COMM_AM2 */


/* GET lptr with explicit counter */
EXTERN INLINE_MODIFIER void __lp_get_ctr(void **lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
	int box = tobox(gPtr);
	void **addr = tolocal(gPtr);

	if (box == MYBOX) {
	  *lPtr = *addr;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __LP_GET_H, (jUIntPointer)addr, 
			 (jUIntPointer)lPtr, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	void **addr = tolocal(gPtr);

	*lPtr = *addr;
})
#endif /* COMM_AM2 */


/* GET gptr with explicit counter */
EXTERN INLINE_MODIFIER void __gp_get_ctr(jGPointer *lPtr, jGPointer gPtr, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
	int box = tobox(gPtr);
	jGPointer *addr = tolocal(gPtr);

	if (box == MYBOX) {
	  *lPtr = *addr;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __GP_GET_H, (jUIntPointer)addr, 
			 (jUIntPointer)lPtr, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jGPointer *addr = tolocal(gPtr);

	*lPtr = *addr;
})
#endif /* COMM_AM2 */




/* ..........................
 *
 * GETs with implicit counter
 *
 * ..........................
 */

EXTERN INLINE_MODIFIER void __b_get(jbyte *lPtr, jGPointer gPtr)
INLINE({__b_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __sh_get(jshort *lPtr, jGPointer gPtr)
INLINE({ __sh_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __i_get(jint *lPtr, jGPointer gPtr)
INLINE({ __i_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __f_get(jfloat *lPtr, jGPointer gPtr)
INLINE({ __f_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __d_get(jdouble *lPtr, jGPointer gPtr)
INLINE({ __d_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __l_get(jlong *lPtr, jGPointer gPtr)
INLINE({ __l_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __lp_get(void **lPtr, jGPointer gPtr)
INLINE({ __lp_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN INLINE_MODIFIER void __gp_get(jGPointer *lPtr, jGPointer gPtr)
INLINE({ __gp_get_ctr(lPtr, gPtr, &numGetOuts); })


     /*
EXTERN INLINE_MODIFIER void __struct_i_get(void *lPtr, jGPointer gPtr)
INLINE({ __i_get_ctr(lPtr, gPtr, &numGetOuts); })

#ifdef HAS_LONGLONG
EXTERN INLINE_MODIFIER void __struct_d_get(void *lPtr, jGPointer gPtr)
INLINE({ __ll_get_ctr(lPtr, gPtr, &numGetOuts); })
#endif
     */

/* ..........................
 *
 * PUTs with explicit counter
 *
 * ..........................
 */

/* Handler on the remote end to write value */
extern void __b_put_handler (void *token, jbyte *addr,  jbyte data, 
			     Counter *ctr);
extern void __sh_put_handler(void *token, jshort *addr, jshort data, 
			     Counter *ctr);
extern void __i_put_handler (void *token, jint *addr,   jint data, 
			     Counter *ctr);
extern void __f_put_handler (void *token, jfloat *addr, jfloat data, 
			     Counter *ctr);

extern void __d_put_handler (void *token, jdouble *addr, jdouble data, 
			     Counter *ctr);
extern void __l_put_handler (void *token, jlong *addr, jlong data, 
			     Counter *ctr);

extern void __lp_put_handler (void *token, void **addr, void *data, 
			      Counter *ctr);

#ifdef PTR64
extern void __gp_put_handler (void *token, jGPointer *addr, 
			      jUIntPointer data0, jUIntPointer data1, 
			      Counter *ctr);
#else
extern void __gp_put_handler (void *token, jGPointer *addr, 
			      jGPointer data,
			      Counter *ctr);
#endif

/* PUT character with explicit counter */
EXTERN INLINE_MODIFIER void __b_put_ctr(jGPointer gPtr, jbyte val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
	int box = tobox(gPtr);
	jbyte *addr = tolocal(gPtr);

	if (box == MYBOX) {
	  *addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __B_PUT_H, (jUIntPointer)addr, 
		      val, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jbyte *addr = tolocal(gPtr);
	  *addr = val;
})
#endif /* COMM_AM2 */

/* PUT short with explicit counter */
EXTERN INLINE_MODIFIER void __sh_put_ctr(jGPointer gPtr, jshort val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
  int box = tobox(gPtr);
	jshort *addr = tolocal(gPtr);

	if (box == MYBOX) {
	  *addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __SH_PUT_H, (jUIntPointer)addr, 
		      val, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jshort *addr = tolocal(gPtr);
	*addr = val;
})
#endif /* COMM_AM2 */

/* PUT int with explicit counter */
EXTERN INLINE_MODIFIER void __i_put_ctr(jGPointer gPtr, jint val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
INLINE({
	int box = tobox(gPtr);
	jint *addr = tolocal(gPtr);

	if (box == MYBOX) {
		*addr = val;
	} else {
		__incr_ctr(ctr);
		
		TI_AM_Request3(CUR_ENDPOINT, box, __I_PUT_H, (jUIntPointer)addr, 
			       val, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jint *addr = tolocal(gPtr);
		*addr = val;
})
#endif /* COMM_AM2 */


/* PUT float with explicit counter */
EXTERN INLINE_MODIFIER void __f_put_ctr(jGPointer gPtr, jfloat val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2 
INLINE({
	int box = tobox(gPtr);
	jfloat *addr = tolocal(gPtr);

	if (box == MYBOX) {
		*addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __F_PUT_H, (jUIntPointer)addr, 
		      val, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	jfloat *addr = tolocal(gPtr);
		*addr = val;
})
#endif /* COMM_AM2 */


EXTERN INLINE_MODIFIER void __d_put_ctr(jGPointer gPtr, jdouble val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
#ifdef PTR64
/* PUT double with explicit counter */
INLINE({
	int box = tobox(gPtr);
	jdouble *addr = tolocal(gPtr);

	if (box == MYBOX) {
		*addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __D_PUT_H, (jUIntPointer)addr, 
			 val, (jUIntPointer)ctr);
	}
})
#else
/* PUT double with explicit counter */
INLINE({
	int box = tobox(gPtr);
	jdouble *addr = tolocal(gPtr);

	if (box == MYBOX) {
		*addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request4(CUR_ENDPOINT, box, __D_PUT_H, (jUIntPointer)addr, 
			 HIGHWORD(*((julong *)&val)), 
			 LOWWORD(*((julong *)&val)), (jUIntPointer)ctr);
	}
})
#endif /* PTR64 */
#else
INLINE({
	jdouble *addr = tolocal(gPtr);
		*addr = val;
})
#endif /* COMM_AM2 */




/* PUT long with explicit counter */
EXTERN INLINE_MODIFIER void __l_put_ctr(jGPointer gPtr, jlong val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
#ifdef PTR64
INLINE({
	int box = tobox(gPtr);
	jlong *addr = tolocal(gPtr);

	if (box == MYBOX) {
		*addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request3(CUR_ENDPOINT, box, __L_PUT_H, (jUIntPointer)addr, 
			 val, (jUIntPointer)ctr);
	}
})
#else
INLINE({
	int box = tobox(gPtr);
	jlong *addr = tolocal(gPtr);

	if (box == MYBOX) {
		*addr = val;
	} else {
	  __incr_ctr(ctr);
	  
	  TI_AM_Request4(CUR_ENDPOINT, box, __L_PUT_H, (jUIntPointer)addr, 
		      HIGHWORD((julong)val), 
		      LOWWORD((julong)val), (jUIntPointer)ctr);
	}
})
#endif /* PTR64 */
#else
INLINE({
	jlong *addr = tolocal(gPtr);
		*addr = val;
})
#endif /* COMM_AM2 */

/* PUT Local Pointer with explicit counter */
EXTERN INLINE_MODIFIER void __lp_put_ctr(jGPointer gPtr, void *val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
     INLINE({
  int box         = tobox(gPtr);
  void **addr = tolocal(gPtr);
     
  if (box == MYBOX) {
    *addr = val;
  } else {
    __incr_ctr(ctr);
    
    TI_AM_Request3(CUR_ENDPOINT, box, __LP_PUT_H, (jUIntPointer)addr, 
		   (jUIntPointer)val, (jUIntPointer)ctr);
  }
})
#else
INLINE({
	void **addr = tolocal(gPtr);
	*addr = val;
})
#endif /* COMM_AM2 */


/* PUT Global Pointer with explicit counter */
EXTERN INLINE_MODIFIER void __gp_put_ctr(jGPointer gPtr, jGPointer val, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
#ifdef PTR64 
     INLINE({
  int box         = tobox(gPtr);
  jGPointer *addr = tolocal(gPtr);
  jUIntPointer data0;
  jUIntPointer data1;
     
  if (box == MYBOX) {
    *addr = val;
  } else {
    data0 = HIGHWORD(val);
    data1 = LOWWORD(val);
    __incr_ctr(ctr);
    
    TI_AM_Request4(CUR_ENDPOINT, box, __GP_PUT_H, (jUIntPointer)addr, 
		   data0, data1, (jUIntPointer)ctr);
  }
})
#else /* PTR64 */
     INLINE({
  int box         = tobox(gPtr);
  jGPointer *addr = tolocal(gPtr);
  Box boxData;
  Process procData;
  void *addrData;

  if (box == MYBOX) {
    *addr = val;
  } else {
    boxData = tobox(val);
    procData = toproc(val);
    addrData = tolocal(val);
    __incr_ctr(ctr);
    
    TI_AM_Request5(CUR_ENDPOINT, box, __GP_PUT_H, (jUIntPointer)addr, 
		   boxData, procData, addrData, (jUIntPointer)ctr);
  }
})
#endif /* PTR64 */
#else
INLINE({
	jGPointer *addr = tolocal(gPtr);
	*addr = val;
})
#endif /* COMM_AM2 */



/* ..........................
 *
 * PUTs with implicit counter
 *
 * ..........................
 */

#if 0
extern void __d_put_implicit_handler(void *token, double *addr,
				     unsigned word0, unsigned word1);

#ifdef HAS_LONGLONG
extern void __ll_put_implicit_handler(void *token, long long *addr,
				      unsigned word0, unsigned word1);
#endif
#endif


EXTERN INLINE_MODIFIER void __b_put(jGPointer gPtr, jbyte val)
INLINE({ __b_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __sh_put(jGPointer gPtr, jshort val)
INLINE({ __sh_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __i_put(jGPointer gPtr, jint val)
INLINE({ __i_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __f_put(jGPointer gPtr, jfloat val)
INLINE({ __f_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __d_put(jGPointer gPtr, jdouble val)
INLINE({ __d_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __l_put(jGPointer gPtr, jlong val)
INLINE({ __l_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __lp_put(jGPointer gPtr, void *val)
INLINE({ __lp_put_ctr(gPtr, val, &numPutOuts); })

EXTERN INLINE_MODIFIER void __gp_put(jGPointer gPtr, jGPointer val)
INLINE({ __gp_put_ctr(gPtr, val, &numPutOuts); })




/* .....
 *
 * READs
 *
 * .....
 */

/* READ character */
EXTERN INLINE_MODIFIER jbyte __b_read(jGPointer gPtr)
INLINE({
	jbyte val;
	Counter ctr=0;
	
	__b_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ short */
EXTERN INLINE_MODIFIER jshort __sh_read(jGPointer gPtr)
INLINE({
	jshort val;
	Counter ctr=0;
	
	__sh_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ int */
EXTERN INLINE_MODIFIER jint __i_read(jGPointer gPtr)
INLINE({
	jint val;
	Counter ctr=0;

	__i_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ float */
EXTERN INLINE_MODIFIER jfloat __f_read(jGPointer gPtr)
INLINE({
	jfloat val;
	Counter ctr=0;

	__f_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ double */
EXTERN INLINE_MODIFIER jdouble __d_read(jGPointer gPtr)
INLINE({
	jdouble val;
	Counter ctr=0;

	__d_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ long long */
EXTERN INLINE_MODIFIER jlong __l_read(jGPointer gPtr)
INLINE({
	jlong val;
	Counter ctr=0;

	__l_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

     /* READ local pointer */
EXTERN INLINE_MODIFIER void * __lp_read(jGPointer gPtr)
INLINE({
	void *val;
	Counter ctr=0;

	__lp_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})


     /* READ global pointer */
EXTERN INLINE_MODIFIER jGPointer __gp_read(jGPointer gPtr)
INLINE({
	jGPointer val;
	Counter ctr=0;

	__gp_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

#if 0

EXTERN	___Struct_i val_i;     /* this is not re-entrant */  

/* READ 4 byte struct */
EXTERN INLINE_MODIFIER ___Struct_i * __struct_i_read(jGPointer gPtr)
INLINE({

	Counter ctr=0;

	__i_get_ctr(&val_i.data, gPtr, &ctr);
	sync_ctr(&ctr);
	return &val_i;
})

#ifdef HAS_LONGLONG
EXTERN 	___Struct_d val_d;   /* this is not re-entrant */

/* READ 8 bytes struct */
EXTERN INLINE_MODIFIER ___Struct_d * __struct_d_read(jGPointer gPtr)
INLINE({

	Counter ctr=0;

	__ll_get_ctr(&val_d.data, gPtr, &ctr);
	sync_ctr(&ctr);
	return &val_d;
})
#endif /* HAS_LONGLONG */

#endif /* 0 */



/* ......
 *
 * WRITEs
 *
 * ......
 */

extern Counter __write_ctr;

#if 0
extern void __d_write_handler (void *token, double *addr, unsigned word0,
							   unsigned word1);
#ifdef HAS_LONGLONG
extern void __ll_write_handler(void *token, long long *addr, unsigned word0,
			       unsigned word1);
#endif /* HAS_LONGLONG */
#endif /* 0 */

extern void __decr_write_ctr_reply_handler(void *token);


/* WRITE character */
EXTERN INLINE_MODIFIER jbyte __b_write(jGPointer gPtr, jbyte val)
INLINE({
	Counter ctr=0;

	__b_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* WRITE short */
EXTERN INLINE_MODIFIER jshort __sh_write(jGPointer gPtr, jshort val)
INLINE({
	Counter ctr=0;

	__sh_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* WRITE int */
EXTERN INLINE_MODIFIER jint __i_write(jGPointer gPtr, jint val)
INLINE({
	Counter ctr=0;

	__i_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* WRITE float */
EXTERN INLINE_MODIFIER jfloat __f_write(jGPointer gPtr, jfloat val)
INLINE({
	Counter ctr=0;

	__f_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;

})

/* WRITE double */
EXTERN INLINE_MODIFIER jdouble __d_write(jGPointer gPtr, jdouble val)
INLINE({
	Counter ctr=0;

	__d_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;

})

/* WRITE long */
EXTERN INLINE_MODIFIER jlong __l_write(jGPointer gPtr, jlong val)
INLINE({
	Counter ctr=0;

	__l_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;

})

/* WRITE local pointer */
EXTERN INLINE_MODIFIER void *__lp_write(jGPointer gPtr, void * val)
INLINE({
	Counter ctr=0;

	__lp_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;

})

/* WRITE global pointer */
EXTERN INLINE_MODIFIER jGPointer __gp_write(jGPointer gPtr, jGPointer val)
INLINE({
	Counter ctr=0;

	__gp_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;

})



/* ................................
 *
 * Ununderstandable garbage... -TvE
 *
 * ................................
 */

/* A wrapper required to return the structure */
EXTERN INLINE_MODIFIER void __read_local(void *lPtr, jGPointer gPtr, int len)
INLINE({
	volatile int flag=0;

	__get_ctr(lPtr, gPtr, len, &flag);
	while (flag != 0) tic_poll();
})

#if 0
/* This function should never be executed. Used
   internally by the compiler but should eventually
   be replaced by a C expression or by __read_local. */
EXTERN INLINE_MODIFIER void * __read (jGPointer p, int size)
INLINE({
        printf ("Tic internal error: please report the bug\n");
	return NULL;
})
#endif /* 0 */



#ifdef PTR64

#define ___GET_HANDLER_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _get_handler(void *token, juint addr0, juint addr1, \
				      juint retAddr0, juint retAddr1,  \
				      juint ctr0, juint ctr1);

#define ___GET_HANDLER(shorttype, type) \
void ___ ## shorttype ## _get_handler(void *token, juint addr0, juint addr1, \
				      juint retAddr0, juint retAddr1,  \
				      juint ctr0, juint ctr1) { \
  type *addr; \
  type *retAddr; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&retAddr, retAddr0); \
  SETLOWWORD((julong *)&retAddr, retAddr1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _get_handler(token, addr, retAddr, ctr); \
} 
#else

#define ___GET_HANDLER_PTYPE(shorttype, type) \
void ___ ## shorttype ## _get_handler(void *token, juint addr,  \
				      juint retAddr, juint ctr);

#define ___GET_HANDLER(shorttype, type) \
void ___ ## shorttype ## _get_handler(void *token, juint addr,  \
				      juint retAddr, juint ctr) {  \
  __ ## shorttype ## _get_handler(token, (type *)addr, (type *)retAddr, \
				  (Counter *)ctr); \
}

#endif  /* PTR64 */
 
___GET_HANDLER_PTYPE(b, jbyte)
___GET_HANDLER_PTYPE(sh, jshort)
___GET_HANDLER_PTYPE(i, jint)
___GET_HANDLER_PTYPE(f, jfloat)
___GET_HANDLER_PTYPE(d, jdouble)
___GET_HANDLER_PTYPE(l, jlong)
___GET_HANDLER_PTYPE(lp, void *)
___GET_HANDLER_PTYPE(gp, jGPointer)







#ifdef PTR64

#define ___GET_REPLY_HANDLER4_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _get_reply_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data, \
						   juint ctr0, juint ctr1);

#define ___GET_REPLY_HANDLER4(shorttype, type) \
void ___ ## shorttype ## _get_reply_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data, \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _get_reply_handler(token, addr, (type)data, ctr); \
} 

#define ___GET_REPLY_HANDLER8_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _get_reply_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data0, juint data1,  \
					    juint ctr0, juint ctr1);

#define ___GET_REPLY_HANDLER8(shorttype, type) \
void ___ ## shorttype ## _get_reply_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data0, juint data1,  \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  type data; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&data, data0); \
  SETLOWWORD((julong *)&data, data1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _get_reply_handler(token, addr, data, ctr); \
}

#define ___GET_REPLY_HANDLER16_PTYPE(shorttype, type, halftype) \
extern void ___ ## shorttype ## _get_reply_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data00, juint data01,  \
					    juint data02, juint data03,  \
					    juint ctr0, juint ctr1);


#define ___GET_REPLY_HANDLER16(shorttype, type, halftype) \
void ___ ## shorttype ## _get_reply_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data00, juint data01,  \
					    juint data02, juint data03,  \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  halftype data0, data1; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&data0, data00); \
  SETLOWWORD((julong *)&data0, data01); \
  SETHIGHWORD((julong *)&data1, data02); \
  SETLOWWORD((julong *)&data1, data03); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _get_reply_handler(token, addr, data0, data1, ctr); \
}

 
#else

#define ___GET_REPLY_HANDLER4_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _get_reply_handler(void *token, juint addr,  \
				      juint data0, juint ctr);

#define ___GET_REPLY_HANDLER4(shorttype, type) \
void ___ ## shorttype ## _get_reply_handler(void *token, juint addr,  \
				      juint data0, juint ctr) {  \
  __ ## shorttype ## _get_reply_handler(token, (type *)addr, (type)data0, \
				  (Counter *)ctr); \
}

#define ___GET_REPLY_HANDLER8_PTYPE(shorttype, type) \
void ___ ## shorttype ## _get_reply_handler(void *token, juint addr,  \
				      juint data0, juint data1, juint ctr);

#define ___GET_REPLY_HANDLER8(shorttype, type) \
void ___ ## shorttype ## _get_reply_handler(void *token, juint addr,  \
				      juint data0, juint data1, juint ctr) {  \
  type data; \
  SETHIGHWORD((julong *)&data, data0); \
  SETLOWWORD((julong *)&data, data1); \
  __ ## shorttype ## _get_reply_handler(token, (type *)addr, data, \
				        (Counter *)ctr); \
}


#endif  /* PTR64 */

___GET_REPLY_HANDLER4_PTYPE(b, jbyte)
___GET_REPLY_HANDLER4_PTYPE(sh, jshort)
___GET_REPLY_HANDLER4_PTYPE(i, jint)
___GET_REPLY_HANDLER4_PTYPE(f, jfloat)
___GET_REPLY_HANDLER8_PTYPE(d, jdouble)
___GET_REPLY_HANDLER8_PTYPE(l, jlong)
#ifdef PTR64
     ___GET_REPLY_HANDLER16_PTYPE(gp, jGPointer, jUIntPointer)
     ___GET_REPLY_HANDLER8_PTYPE(lp, void *)
#else
     ___GET_REPLY_HANDLER8_PTYPE(gp, jGPointer)
     ___GET_REPLY_HANDLER4_PTYPE(lp, void *)
#endif







#ifdef PTR64

#define ___PUT_HANDLER4_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _put_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data, \
					    juint ctr0, juint ctr1);


#define ___PUT_HANDLER4(shorttype, type) \
void ___ ## shorttype ## _put_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data, \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _put_handler(token, addr, (type)data, ctr); \
} 

#define ___PUT_HANDLER8_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _put_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data0, juint data1,  \
					    juint ctr0, juint ctr1);

#define ___PUT_HANDLER8(shorttype, type) \
void ___ ## shorttype ## _put_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data0, juint data1,  \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  type data; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&data, data0); \
  SETLOWWORD((julong *)&data, data1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _put_handler(token, addr, data, ctr); \
}

#define ___PUT_HANDLER16(shorttype, type, halftype) \
extern void ___ ## shorttype ## _put_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data00, juint data01,  \
					    juint data02, juint data03,  \
					    juint ctr0, juint ctr1);

#define ___PUT_HANDLER16(shorttype, type, halftype) \
void ___ ## shorttype ## _put_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data00, juint data01,  \
					    juint data02, juint data03,  \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  halftype data0, data1; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&data0, data00); \
  SETLOWWORD((julong *)&data0, data01); \
  SETHIGHWORD((julong *)&data1, data02); \
  SETLOWWORD((julong *)&data1, data03); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _put_handler(token, addr, data0, data1, ctr); \
}

 
#else

#define ___PUT_HANDLER4_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _put_handler(void *token, juint addr,  \
				      juint data0, juint ctr);

#define ___PUT_HANDLER4(shorttype, type) \
void ___ ## shorttype ## _put_handler(void *token, juint addr,  \
				      juint data0, juint ctr) {  \
  __ ## shorttype ## _put_handler(token, (type *)addr, (type)data0, \
				  (Counter *)ctr); \
}

#define ___PUT_HANDLER8_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _put_handler(void *token, juint addr,  \
				      juint data0, juint data1, juint ctr);

#define ___PUT_HANDLER8(shorttype, type) \
void ___ ## shorttype ## _put_handler(void *token, juint addr,  \
				      juint data0, juint data1, juint ctr) {  \
  type data; \
  SETHIGHWORD((julong *)&data, data0); \
  SETLOWWORD((julong *)&data, data1); \
  __ ## shorttype ## _put_handler(token, (type *)addr, data, \
				        (Counter *)ctr); \
}


#endif  /* PTR64 */

___PUT_HANDLER4_PTYPE(b, jbyte)
___PUT_HANDLER4_PTYPE(sh, jshort)
___PUT_HANDLER4_PTYPE(i, jint)
___PUT_HANDLER4_PTYPE(f, jfloat)
___PUT_HANDLER8_PTYPE(d, jdouble)
___PUT_HANDLER8_PTYPE(l, jlong)
#ifdef PTR64
     ___PUT_HANDLER8_PTYPE(lp, void *)
#else
     ___PUT_HANDLER4_PTYPE(lp, void *)
#endif

#ifdef PTR64
     ___PUT_HANDLER16_PTYPE(gp, jGPointer, jUIntPointer)
#else

void ___gp_put_handler(void *token, juint addr, juint box, juint proc, juint addrData,
		       juint ctr);

#endif




#endif /* __MEMORY_H */





