#include <assert.h>
#include <tic.h>

/* TODO: the AM message arguments in this file will need to be fixed up 
 *  if we ever migrate to a 64-bit AM implementation
 */

/**************************
 * INITIALIZATION/STATICS *
 *************************/

#ifdef COMM_AM2
static handler_t packHandlerIdx, unpackAllHandlerIdx, unpackOnlyHandlerIdx;
static handler_t packReplyHandlerIdx, unpackReplyHandlerIdx;
static handler_t deleteHandlerIdx, nullReplyHandlerIdx, allocHandlerIdx,
                 allocReplyHandlerIdx;

typedef void *(*pack_method_t)(void *, int *, void *);
typedef void *(*unpack_method_t)(void *, void *);

static void packHandler(void *token, void *copy_desc, int copy_desc_size,
			int _packMethod, int remoteBuffer, int pack_info_ptr, 
			int dummy3);
static void packReplyHandler(void *token, void *array_data, 
			     int array_data_size, int localBuffer,
			     int remoteBuf, int dataSz,
                             int pack_info_ptr);
static void unpackAllHandler(void *token, void *packedArrayData, int nBytes, 
		             int _unpackMethod, int copyDescSize, int unpack_spin_ptr, 
		             int dummy3);
static void unpackOnlyHandler(void *token, void *copyDesc, int copyDescSize,
			      int bufAddr, int _unpackMethod, int unpack_spin_ptr,
			      int dummy3);
static void unpackReplyHandler(void *token, int unpack_spin_ptr, 
                               int dummy1, int dummy2, int dummy3);
static void deleteHandler(void *token, int addr, int dummy1, int dummy2,
			  int dummy3);
static void allocHandler(void *token, int size, int destptr, int dummy2,
			  int dummy3);
static void allocReplyHandler(void *token, int remAddr, int destptr, 
			      int dummy2, int dummy3);
static void nullReplyHandler(void *token);

void bulkSetupHandlers() {
  if (COMM_GetBParallelDegree() > 1) {
    /* choose any available handler index for these handlers */
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &packHandlerIdx, packHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackAllHandlerIdx, unpackAllHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackOnlyHandlerIdx, unpackOnlyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &packReplyHandlerIdx, packReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackReplyHandlerIdx, unpackReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &deleteHandlerIdx, deleteHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &allocHandlerIdx, allocHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &allocReplyHandlerIdx, 
			     allocReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &nullReplyHandlerIdx, nullReplyHandler));
  }
}

/***************************
 * Miscellaneous handlers. *
 ***************************/

static void nullReplyHandler(void *token) {
  return;
}

static void deleteHandler(void *token, int addr, int dummy1, int dummy2,
			  int dummy3) {
  free((void *)addr);
  TI_AM_Reply0(token, nullReplyHandlerIdx);
}

/* Size is in bytes. */
static void allocHandler(void *token, int size, int destptr, int dummy2,
			  int dummy3) {
  void *buf = (void *)malloc(size);
  TI_AM_Reply4(token, allocReplyHandlerIdx, (int)buf, destptr, 0, 0);
}

static void allocReplyHandler(void *token, int buf, int destptr, int dummy2,
			      int dummy3) {
  void ** premoteAllocBuf = (void **)destptr;
  (*premoteAllocBuf) = (void *)buf;
}

/************************************
 * Get an array from a remote node. *
 ***********************************/

/* Call _packMethod to pack the array with copy_desc as its "copy descriptor",
   returning the packed data to remote address remoteBuffer. */
static void packHandler(void *token, void *copy_desc, int copy_desc_size,
			int _packMethod, int remoteBuffer, int pack_info_ptr, 
			int dummy3) {
  int array_data_len;
  void *array_data;
  pack_method_t packMethod = (pack_method_t)_packMethod;
  /* A NULL buffer pointer tells the packer to allocate its own buffer. */
  array_data = (*packMethod)(copy_desc, &array_data_len, NULL);

  if (array_data_len <= AM_MaxMedium()) {
    /* Common, fast case. Send back the packed data with the reply. */
    TI_AM_ReplyI4(token, packReplyHandlerIdx, array_data, 
		       array_data_len, remoteBuffer, 0, 0, pack_info_ptr);
    free(array_data);
  }
  else {
    /* Slower case. Why doesn't the darned AM_ReplyXfer() work? */
    TI_AM_ReplyI4(token, packReplyHandlerIdx, array_data,
		       AM_MaxMedium(), remoteBuffer, (int)array_data, 
		       array_data_len, pack_info_ptr);
  }
}

/* The reply handler needs to pass this information back to the caller. */
typedef struct {
  volatile int pack_spin; /* 0 while waiting, 1 if completed, 2 if using fragmentation */
  void * volatile remoteBuffer;  
  volatile int dataSize;  
} pack_return_info_t;

/* Sets the spin flag to 1 if the array fit in the message reply and we're
   done; otherwise it sets the spinlock variable to be the remote address 
   of the packed data, which must be sent over with a libtic bulk 
   transfer. */
static void packReplyHandler(void *token, void *array_data, 
			     int array_data_size, int localBuffer,
			     int remoteBuf, int dataSz,
                             int pack_info_ptr) {
  int moreWaiting = (remoteBuf != 0 && dataSz != 0);
  pack_return_info_t* pinfo = (pack_return_info_t*)pack_info_ptr;
  memcpy((void *)localBuffer, array_data, array_data_size);
  if (!moreWaiting) { /* got it all in a single chunk */
    pinfo->pack_spin = 1;
  }
  else { /* need more chunks */
    assert(array_data_size == AM_MaxMedium()); /* first chunk always max medium in size */
    pinfo->remoteBuffer = (jbyte *)remoteBuf;
    pinfo->dataSize = dataSz;
    pinfo->pack_spin = 2;
  }
}

void get_array(void *pack_method, void *copy_desc, int copy_desc_size, 
	       int tgt_node, void *buffer) {
  Box tgt_box = COMM_GetBoxNumberForProcNumber(tgt_node);
  pack_return_info_t info;
  info.pack_spin = 0;
  /* A copy_desc is an object that contains the array descriptor. */
  TI_AM_RequestI4(__sc_endpoint, tgt_box, packHandlerIdx, 
		       copy_desc, copy_desc_size, (int)pack_method, 
		       (int)buffer, (int)&info, 0);
  while (!info.pack_spin) { tic_poll(); }
  /* If the data was too big to be sent in the reply, use a Tic call
     to do the job. */
  if (info.pack_spin != 1) {
    /* pull the data to use using a bulk copy (first fragment has already been copied) */
    jGPointer remoteBufPtr = toglobal(tgt_node, ((jbyte *)info.remoteBuffer) + AM_MaxMedium());
    bulk_read((void *)((int)buffer+AM_MaxMedium()), remoteBufPtr, info.dataSize - AM_MaxMedium());
    /* now tell remote to delete temp space */
    TI_AM_Request4(__sc_endpoint, tgt_box, deleteHandlerIdx, 
			(int)info.remoteBuffer, 0, 0, 0);
  }
}

/************************************
 * Write an array to a remote node. *
 ***********************************/

/* The fastest way of doing things....take the data given and unpack it. */
static void unpackAllHandler(void *token, void *packedArrayData, int nBytes, 
		   int _unpackMethod, int copyDescSize, int unpack_spin_ptr, 
		   int dummy3) {
 void *temp = NULL;
 if (((int)packedArrayData) % 8 != 0) {
   /* it seems that some crappy AM implementations (ahem.. the NOW)
    * fail to double-word align the medium-sized message buffer
    * fix that problem if it occurs
    */
   temp = malloc(nBytes+8);
   memcpy(temp, packedArrayData, nBytes);
   packedArrayData = temp;
 } 
 {
   unpack_method_t unpackMethod = (unpack_method_t)_unpackMethod;
   void *copyDesc = packedArrayData;
   void *arrayData = (void *)((int)packedArrayData + copyDescSize);
   (*unpackMethod)(copyDesc, arrayData);
   TI_AM_Reply4(token, unpackReplyHandlerIdx, unpack_spin_ptr, 0, 0, 0);
 }
 if (temp) free(temp);
}

static void unpackReplyHandler(void *token, int unpack_spin_ptr, 
                               int dummy1, int dummy2, int dummy3) {
  int *punpack_spin = (int *)unpack_spin_ptr;
  *(punpack_spin) = 1;
}

/* The data was provided by previous calls; just unpack it with the given
   descriptor. */
static void unpackOnlyHandler(void *token, void *copyDesc, int copyDescSize,
			      int bufAddr, int _unpackMethod, int unpack_spin_ptr,
			      int dummy3) {
  unpack_method_t unpackMethod = (unpack_method_t)_unpackMethod;
  (*unpackMethod)(copyDesc, (void *)bufAddr);
  free((void *)bufAddr);
  TI_AM_Reply4(token, unpackReplyHandlerIdx, unpack_spin_ptr, 0, 0, 0);
}
  

/* Send a contiguous array of data to a node, and have it get unpacked into
   a Titanium array. */
void put_array(void *unpack_method, void *copy_desc, int copy_desc_size,
	       void *array_data, int array_data_size, int tgt_node) {
  void *data;
  /* ensure double-word alignment for array data */
  int copy_desc_size_padded = ((copy_desc_size-1)/8 + 1) * 8; 
  int data_size = copy_desc_size_padded + array_data_size;
  Box tgt_box = COMM_GetBoxNumberForProcNumber(tgt_node);
  volatile int unpack_spin = 0;

  /* Fast(er), hopefully common case. */
  if (data_size <= AM_MaxMedium()) {
    /* The remote end needs the descriptor and the array data. Since the
       data is small, copying the array data will be faster than
       sending two messages. */
    data = (void *) malloc(data_size);
    assert(data && ((int)data) % 8 == 0 && data_size % 8 == 0);
    memcpy(data, copy_desc, copy_desc_size);
    /* All math is done in bytes. sizeof(void*) should be irrelevant. */
    memcpy((void *)((int)data + copy_desc_size_padded), array_data, array_data_size);
    TI_AM_RequestI4(__sc_endpoint, tgt_box, unpackAllHandlerIdx, 
			 data, data_size, (int)unpack_method, 
			 copy_desc_size_padded, (int)&unpack_spin, 0);
    while (!unpack_spin) { tic_poll(); }
    free(data);
  }
  else { /* Slow case. */
    jGPointer remotePtr;
    /* Allocate a buffer to hold the array data on the remote side. */
    volatile void *remoteAllocBuf = NULL;
    TI_AM_Request4(__sc_endpoint, tgt_box, allocHandlerIdx, 
			array_data_size, (int)&remoteAllocBuf, 0, 0);
    while (!remoteAllocBuf) { tic_poll(); }
    /* Transfer the data to the buffer with libtic. */
    remotePtr = toglobal(tgt_node, (void *)remoteAllocBuf);
    bulk_write(remotePtr, array_data, array_data_size);
    /* Tell the remote side to unpack the data. */
    TI_AM_RequestI4(__sc_endpoint, tgt_box, unpackOnlyHandlerIdx,
			 copy_desc, copy_desc_size, (int)remoteAllocBuf, 
			 (int)unpack_method, (int)&unpack_spin, 0);
    while (!unpack_spin) { tic_poll(); }
  }
}

#endif /* COMM_AM2 */
