/*									tab:4
 *
 * barrier.c - Code to implement Tic barrier
 *
 * "Copyright (c) 1996 and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Andrea C. Dusseau
 * Version:			3
 * Creation Date:	Mon Feb 12 23:26:01 1996
 * Filename:		barrier.c
 * History:
 */

#include <stdio.h>
#include <sys/time.h>
#if HAVE_SCHED_YIELD
#include <sched.h>
#endif
#include <errno.h>
#include <assert.h>

#include <tic.h>

#if HAVE_ALLOCA_H
#include <alloca.h>
#endif

#define DEBUG_BARRIERS 0
#define DEBUG_BARRIERS_VERBOSE 0
#if DEBUG_BARRIERS
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif


typedef struct {
  __tic_lock_t lock; /* these three belong on a cache line together */
  volatile int counter;
  tic_simplemonitor_t mon;

  char _pad1[256]; /* make sure flag is on a cache line by itself */
  volatile int flag;
  char _pad2[256]; /* make sure flag is on a cache line by itself */

} tic_barrier_t;

static volatile tic_barrier_t local_barrier_object __attribute__((aligned));

__tic_lock_t	Barrier_mutex;
static Counter up_buf[2];
static Counter down_buf;

void barrier_init()
{
  /* CAUTION: this routine executes very early during startup,
      before any of the system-wide parameters have been set
      don't access rely on the values of any globals or parameter-returning 
      functions (e.g. MYBOXPROCS, etc.) in here */
  int i;

  up_buf[0] = 0;
  up_buf[1] = 0;
  down_buf  = 0;

  /* these are OK */
  for (i = 0; i < COMM_GetMyBParallelDegree(); i++) {
    COMM_GetHisBoxProcess(i)->barrier_phase = 0;
  }

  /* must use __tic_lock_initializer because __tic_lock_init may reference MYBOXPROCS */
  Barrier_mutex =  __tic_lock_initializer;
  local_barrier_object.lock =  __tic_lock_initializer;
  simplemonitor_init(((tic_simplemonitor_t *)&(local_barrier_object.mon)));

  local_barrier_object.counter = 0;
  local_barrier_object.flag = 0;
}

#ifdef COMM_AM2
void __barrier_incr_handler(void *token, int *val)
{
  __tic_lock(&Barrier_mutex);
  (*val)++;
  __tic_unlock(&Barrier_mutex);
  TI_AM_Reply0(token, __NULL_RH);
}
#endif /* COMM_AM2 */

#if DEBUG_BARRIERS
volatile int debug_localbarriercount = 0;
#endif



#if 1
/*------------------------------------------------------------------------------------*/
/* new barrier algorithm 
 * written & tuned by Dan Bonachea <bonachea@cs.berkeley.edu>
 * key design issues (for the "impolite" high-performance version): 
 *   avoid expensive pthread mutex locking
 *   design for a linear number of cache invalidations for good scaling
 *   avoid any false sharing of cache lines
 *   pipeline the reads of remote data for coherence implementations that allow
 *    multiple outstanding remote reads (like the Origin 2000)
 */
void local_barrier() {
  #if DEBUG_BARRIERS_VERBOSE 
    fprintf(stderr, "P%i: Barrier %i Enter\n", MYBOXPROC, debug_localbarriercount);
    fflush(stderr);
  #endif

  /* ensure that all previous writes have been globally completed */
  tic_local_memory_sync();

  if (politep) { /* "polite" barrier (low-performance) */
    const int numboxprocs = MYBOXPROCS;
    tic_simplemonitor_t * const mon = (tic_simplemonitor_t *)&(local_barrier_object.mon);
    int mycount;

    __tic_lock(&(mon->lock));
    mycount = local_barrier_object.counter++;
    ASSERT(mycount >= 0 && mycount < numboxprocs);
    if (mycount == numboxprocs-1) {
      #if DEBUG_BARRIERS
       fprintf(stderr, "--- Barrier %i ---\n", debug_localbarriercount++);
       fflush(stderr);
      #endif
      local_barrier_object.counter = 0; 
      __tic_mon_broadcast(mon);
    }
    else __tic_mon_wait(mon);
    __tic_unlock(&(mon->lock));

    #if DEBUG_BARRIERS_VERBOSE 
      fprintf(stderr, "P%i: Barrier %i Exit\n", MYBOXPROC, debug_localbarriercount-1);
      fflush(stderr);
    #endif
    return;
  }
  else { /* "impolite" high-performance barrier (assumes thread has CPU to itself) */
    processInfo * const pi = COMM_GetMyProcess();
    const int myboxproc = MYBOXPROC;
    register const int oldphase = pi->barrier_phase;
    const int newphase = !oldphase;
    register int i;

    pi->barrier_phase = newphase; /* signal arrival */

    if (myboxproc == 0) {
      const int numboxprocs = MYBOXPROCS;
      register int n = numboxprocs - 1;
      #if HAVE_ALLOCA
        /* use alloca to guarantee thread-affinity of this memory */
        register volatile int ** const addr = (volatile int **)alloca(numboxprocs*sizeof(int*));
        register int * const val = (int *)alloca(numboxprocs*sizeof(int));
      #else
        static volatile int ** addr = NULL;
        static int * val = NULL;
        if (!addr) {
          addr = (volatile int **)malloc(numboxprocs*sizeof(int*));
          val = (int *)malloc(numboxprocs*sizeof(int));
        }
      #endif

      for (i = 1; i < numboxprocs; i++) {
        register processInfo const *rpi = COMM_GetHisBoxProcess(i);
        addr[i-1] = (volatile int *)&(rpi->barrier_phase);
        val[i-1] = rpi->barrier_phase; /* pipelined read */
      }
      while (n) { /* n is number of outstanding procs */
        for (i = 0; i < n; ) {   
          if (val[i] == oldphase) {
            val[i] = *addr[i]; /* pipelined read */
            i++;
          }
          else {
            n--;
            addr[i] = addr[n];   
            val[i] = val[n];
          }
        }
      }

      #if DEBUG_BARRIERS_VERBOSE 
        fprintf(stderr, "--- Barrier %i ---\n", debug_localbarriercount++);
        fflush(stderr);
      #endif

      local_barrier_object.flag = newphase; /* signal completion */

    } else { /* procs other than 0 */
      /* Might as well spin, no one else is running on this proc anyway */
      while (local_barrier_object.flag == oldphase);
    }
  
    #if DEBUG_BARRIERS_VERBOSE 
      fprintf(stderr, "P%i: Barrier %i Exit\n", MYBOXPROC, debug_localbarriercount-1);
      fflush(stderr);
    #endif
    return;
  }
}

#else 
/*------------------------------------------------------------------------------------*/

int __os_sleep(long secs, long usecs);
void __os_yield(long usecs);

#if HAVE_SCHED_YIELD
  #define THREAD_YIELD() sched_yield();
#else
  #define THREAD_YIELD() __os_yield(0);
#endif

#define POLITE_SPINBLOCK_UNTIL(cond) do {   \
  register int usecs = 10;                  \
  register int done = 0;                    \
  while (!done) {                           \
    register int nspins;                    \
    for(nspins = 25; nspins>0; --nspins) {  \
      if (cond) { done = 1; break; }        \
      THREAD_YIELD();                       \
    }                                       \
    __os_yield(usecs);                      \
    if ((usecs <<=1) > 10000) usecs = 10000;\
  }                                         \
} while (0)

/* old barrier algorithm */
void local_barrier() {
  int mycount, nspins; 
  unsigned long usecs;
  int phase;
  processInfo *pi = COMM_GetMyProcess(); 

  #if DEBUG_BARRIERS_VERBOSE 
    int debug_mybarriercount =  debug_localbarriercount;
    fprintf(stderr, "P%i: Barrier %i Enter\n", MYBOXPROC, debug_mybarriercount);
    fflush(stderr);
  #endif

  phase = !(pi->barrier_phase);
  pi->barrier_phase = phase;

  ASSERT(phase != local_barrier_object.flag);

  __tic_lock((__tic_lock_t*)&local_barrier_object.lock);
  mycount = local_barrier_object.counter++;
  ASSERT(mycount < MYBOXPROCS);

  /* Done yet? */
  if (mycount == MYBOXPROCS-1) {
    #if DEBUG_BARRIERS
     fprintf(stderr, "--- Barrier %i ---\n", debug_localbarriercount);
     fflush(stderr);
     debug_localbarriercount++;
    #endif
    /* it's crucial for the counter reset to hit memory before the flag write */
    local_barrier_object.counter = 0; 
    __tic_unlock((__tic_lock_t*)&local_barrier_object.lock);
    local_barrier_object.flag = phase; /* signal completion */
    #if DEBUG_BARRIERS_VERBOSE
      fprintf(stderr, "P%i: Barrier %i Exit\n", MYBOXPROC, debug_mybarriercount);
      fflush(stderr);
    #endif
    return;
  }
  else {
    __tic_unlock((__tic_lock_t*)&local_barrier_object.lock);

    if (!politep) {
      /* Might as well spin, no one else is running anyway */
      while(local_barrier_object.flag != phase); 
    }
    else {
    /* More than one process per processor
       check flag and yield if not set 
       After 25 times, sleep for 1000 usecs
       If this still isn't good, sleep longer. */
      POLITE_SPINBLOCK_UNTIL(local_barrier_object.flag == phase);
    }

    #if DEBUG_BARRIERS_VERBOSE 
      fprintf(stderr, "P%i: Barrier %i Exit\n", MYBOXPROC, debug_mybarriercount);
      fflush(stderr);
    #endif
    return;
  }
}
#endif
/*------------------------------------------------------------------------------------*/

void barrier_up()
{

  /* If there's only one box, just do a local barrier */
  if (COMM_GetBParallelDegree() == 1) {
#ifndef MEMORY_SHARED
    return;
#else
    /* Only need to do something if there are other threads here */
    if (MYBOXPROCS > 1) local_barrier();
    return;
#endif /* NO_SMP */
  }
  
  /* If we're on an SMP, run a local barrier */
  if (MYBOXPROCS > 1) {
    /* MYBOXPROC == 0 does local barrier BEFORE global barrier in barrier_up()
       to prevent remote processors from continuing before all local threads arrive */
    local_barrier();
    if (MYBOXPROC > 0) return;
  }

#ifdef COMM_AM2
{
  /* Only MYBOXPROC = 0 on each box gets to run this part */
  const int parent = (MYBOX - 1) / 2;
  const int odd_child = 2 * MYBOX + 1;
  const int even_child = 2 * MYBOX + 2;
  const int parity = MYBOX & 1;

  if (even_child < BOXES) {
    while (up_buf[0] == 0) {
      tic_poll();
    }
  } 
  if (odd_child < BOXES) {
    while (up_buf[1] == 0) {
      tic_poll();
    }
  }

  if (MYBOX > 0) {
    ADVANCE_SEND_ENDPOINT;
    TI_AM_Request1(CUR_ENDPOINT, parent, __BARRIER_INCR_H, 
			(jUIntPointer)&up_buf[parity]);
  }

  /* not zero */
  __tic_lock(&Barrier_mutex);
  up_buf[0] = up_buf[1] = 0;
  __tic_unlock(&Barrier_mutex);
}
#endif /* AMII */
}

void barrier_down()
{

  if (COMM_GetBParallelDegree() == 1) {
#ifndef MEMORY_SHARED
    return;
#else
    /* Only need to do something if there are other threads here */
    if (MYBOXPROCS > 1) local_barrier();
    return;
#endif /* NO_SMP */
  }
  
  if (MYBOXPROC > 0) {
    local_barrier();
    return;
  }

#ifdef COMM_AM2
{
  /* Only MYBOXPROC = 0 on each box runs this part */
  const int left = 2 * MYBOX + 1;
  const int right = 2 * MYBOX + 2;

  if (MYBOX > 0) {
    while (down_buf == 0) {
      tic_poll();
    }
  }

  if (left < BOXES) {
    ADVANCE_SEND_ENDPOINT;
    TI_AM_Request1(CUR_ENDPOINT, left, __BARRIER_INCR_H, 
			(jUIntPointer)&down_buf);
  }
  if (right < BOXES) {
    ADVANCE_SEND_ENDPOINT;
    TI_AM_Request1(CUR_ENDPOINT, right, __BARRIER_INCR_H, 
			(jUIntPointer)&down_buf);
  }

  __tic_lock(&Barrier_mutex);
  down_buf = 0;
  __tic_unlock(&Barrier_mutex);
}
#endif /* AMII */
  
  if (MYBOXPROCS > 1) {
    local_barrier();
    return;
  }
} 



/*
 * __os_sleep --
 *	Yield the processor for a period of time.
 *
 * PUBLIC: int __os_sleep __P((u_long, u_long));
 */
int
__os_sleep(long secs, long usecs)
{
#if 1 
  /* the small-interval performance of this totally sucks on Linux,
   * where the minimum interval is 10 ms (WAY too long to be useful for much of anything)
   * the same is apparently true for other OS's (IRIX, etc.)
   * why doesn't anybody implement high-granularity timers?
   */
  struct timeval t;

  /* Don't require that the values be normalized. */
  for (; usecs >= 1000000; ++secs, usecs -= 1000000) {}

  /*
   * It's important that we yield the processor here so that other
   * processes or threads are permitted to run.
   */
  t.tv_sec = secs;
  t.tv_usec = usecs;
  return (select(0, NULL, NULL, NULL, &t) == -1 ? errno : 0);
#elif 0
  /* unfortunately, these are no better */
  struct timespec t;

  /* Don't require that the values be normalized. */
  for (; usecs >= 1000000; ++secs, usecs -= 1000000) {}

  t.tv_sec = secs;
  t.tv_nsec = usecs*1000;
  return (nanosleep(&t, NULL) == -1 ? errno : 0);
#elif 0
  return (poll(NULL, 0, (int)(secs*1000.0 + usecs/1000.0)) == -1 ? errno : 0);
#else
  usleep((int)(secs/1000000.0 + usecs));
  return 0;
#endif
}

/*
 * __os_yield --
 *	Yield the processor.
 *
 * PUBLIC: void __os_yield __P((u_long));
 */
void
__os_yield(long usecs)
{
	__os_sleep(0, usecs);
}


